""" MLP module w/ dropout and configurable activation layer

Hacked together by / Copyright 2020 Ross Wightman
"""

from torch import nn as nn

from .layer_helpers import to_2tuple


class Mlp(nn.Module):
    """MLP as used in Vision Transformer, MLP-Mixer and related networks"""

    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        act_layer=nn.GELU,
        bias=True,
        drop=0.0,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        bias = to_2tuple(bias)
        drop_probs = to_2tuple(drop)

        self.fc1 = nn.Linear(in_features, hidden_features, bias=bias[0])
        self.act = act_layer()
        self.drop1 = nn.Dropout(drop_probs[0])
        self.fc2 = nn.Linear(hidden_features, out_features, bias=bias[1])
        self.drop2 = nn.Dropout(drop_probs[1])

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop1(x)
        x = self.fc2(x)
        x = self.drop2(x)
        return x


class ConvMlp(nn.Module):
    """MLP using 1x1 convs that keeps spatial dims"""

    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        act_layer=nn.ReLU,
        norm_layer=None,
        bias=True,
        drop=0.0,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        bias = to_2tuple(bias)

        self.fc1 = nn.Conv2d(in_features, hidden_features, kernel_size=1, bias=bias[0])
        self.norm = norm_layer(hidden_features) if norm_layer else nn.Identity()
        self.act = act_layer()
        self.drop = nn.Dropout(drop)
        self.fc2 = nn.Conv2d(hidden_features, out_features, kernel_size=1, bias=bias[1])

    def forward(self, x):
        x = self.fc1(x)
        x = self.norm(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        return x



"""
Portions of this code copyright 2017, Clement Pinard
"""

import numpy as np

try:
    from spatial_correlation_sampler import SpatialCorrelationSampler
except ModuleNotFoundError:
    from ptlflow.utils.correlation import (
        IterSpatialCorrelationSampler as SpatialCorrelationSampler,
    )
import torch
import torch.nn as nn

from ptlflow.utils.registry import register_model, trainable
from ..base_model.base_model import BaseModel
from ..flownet.losses import MultiScale


def conv(in_planes, out_planes, kernel_size=3, stride=1, padding=1, dilation=1):
    return nn.Sequential(
        nn.Conv2d(
            in_planes,
            out_planes,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            bias=True,
        ),
        nn.LeakyReLU(0.1),
    )


def predict_flow(in_planes):
    return nn.Conv2d(in_planes, 2, kernel_size=3, stride=1, padding=1, bias=True)


def deconv(in_planes, out_planes, kernel_size=4, stride=2, padding=1):
    return nn.ConvTranspose2d(
        in_planes, out_planes, kernel_size, stride, padding, bias=True
    )


class PWCNet(BaseModel):
    pretrained_checkpoints = {
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/pwcnet-things-6a2e540b.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/pwcnet-sintel-533815e5.ckpt",
    }

    def __init__(
        self,
        div_flow: float = 20.0,
        md: int = 4,
        loss_start_scale: int = 4,
        loss_num_scales: int = 5,
        loss_base_weight: float = 0.32,
        loss_norm: str = "L2",
        **kwargs,
    ):
        super(PWCNet, self).__init__(
            loss_fn=MultiScale(
                startScale=loss_start_scale,
                numScales=loss_num_scales,
                l_weight=loss_base_weight,
                norm=loss_norm,
            ),
            output_stride=64,
            **kwargs,
        )

        self.div_flow = div_flow
        self.md = md

        self.conv1a = conv(3, 16, kernel_size=3, stride=2)
        self.conv1aa = conv(16, 16, kernel_size=3, stride=1)
        self.conv1b = conv(16, 16, kernel_size=3, stride=1)
        self.conv2a = conv(16, 32, kernel_size=3, stride=2)
        self.conv2aa = conv(32, 32, kernel_size=3, stride=1)
        self.conv2b = conv(32, 32, kernel_size=3, stride=1)
        self.conv3a = conv(32, 64, kernel_size=3, stride=2)
        self.conv3aa = conv(64, 64, kernel_size=3, stride=1)
        self.conv3b = conv(64, 64, kernel_size=3, stride=1)
        self.conv4a = conv(64, 96, kernel_size=3, stride=2)
        self.conv4aa = conv(96, 96, kernel_size=3, stride=1)
        self.conv4b = conv(96, 96, kernel_size=3, stride=1)
        self.conv5a = conv(96, 128, kernel_size=3, stride=2)
        self.conv5aa = conv(128, 128, kernel_size=3, stride=1)
        self.conv5b = conv(128, 128, kernel_size=3, stride=1)
        self.conv6aa = conv(128, 196, kernel_size=3, stride=2)
        self.conv6a = conv(196, 196, kernel_size=3, stride=1)
        self.conv6b = conv(196, 196, kernel_size=3, stride=1)

        self.leakyRELU = nn.LeakyReLU(0.1)

        self.corr = SpatialCorrelationSampler(
            kernel_size=1, patch_size=2 * self.md + 1, padding=0
        )

        nd = (2 * self.md + 1) ** 2
        dd = np.cumsum([128, 128, 96, 64, 32])

        od = nd
        self.conv6_0 = conv(od, 128, kernel_size=3, stride=1)
        self.conv6_1 = conv(od + dd[0], 128, kernel_size=3, stride=1)
        self.conv6_2 = conv(od + dd[1], 96, kernel_size=3, stride=1)
        self.conv6_3 = conv(od + dd[2], 64, kernel_size=3, stride=1)
        self.conv6_4 = conv(od + dd[3], 32, kernel_size=3, stride=1)
        self.predict_flow6 = predict_flow(od + dd[4])
        self.deconv6 = deconv(2, 2, kernel_size=4, stride=2, padding=1)
        self.upfeat6 = deconv(od + dd[4], 2, kernel_size=4, stride=2, padding=1)

        od = nd + 128 + 4
        self.conv5_0 = conv(od, 128, kernel_size=3, stride=1)
        self.conv5_1 = conv(od + dd[0], 128, kernel_size=3, stride=1)
        self.conv5_2 = conv(od + dd[1], 96, kernel_size=3, stride=1)
        self.conv5_3 = conv(od + dd[2], 64, kernel_size=3, stride=1)
        self.conv5_4 = conv(od + dd[3], 32, kernel_size=3, stride=1)
        self.predict_flow5 = predict_flow(od + dd[4])
        self.deconv5 = deconv(2, 2, kernel_size=4, stride=2, padding=1)
        self.upfeat5 = deconv(od + dd[4], 2, kernel_size=4, stride=2, padding=1)

        od = nd + 96 + 4
        self.conv4_0 = conv(od, 128, kernel_size=3, stride=1)
        self.conv4_1 = conv(od + dd[0], 128, kernel_size=3, stride=1)
        self.conv4_2 = conv(od + dd[1], 96, kernel_size=3, stride=1)
        self.conv4_3 = conv(od + dd[2], 64, kernel_size=3, stride=1)
        self.conv4_4 = conv(od + dd[3], 32, kernel_size=3, stride=1)
        self.predict_flow4 = predict_flow(od + dd[4])
        self.deconv4 = deconv(2, 2, kernel_size=4, stride=2, padding=1)
        self.upfeat4 = deconv(od + dd[4], 2, kernel_size=4, stride=2, padding=1)

        od = nd + 64 + 4
        self.conv3_0 = conv(od, 128, kernel_size=3, stride=1)
        self.conv3_1 = conv(od + dd[0], 128, kernel_size=3, stride=1)
        self.conv3_2 = conv(od + dd[1], 96, kernel_size=3, stride=1)
        self.conv3_3 = conv(od + dd[2], 64, kernel_size=3, stride=1)
        self.conv3_4 = conv(od + dd[3], 32, kernel_size=3, stride=1)
        self.predict_flow3 = predict_flow(od + dd[4])
        self.deconv3 = deconv(2, 2, kernel_size=4, stride=2, padding=1)
        self.upfeat3 = deconv(od + dd[4], 2, kernel_size=4, stride=2, padding=1)

        od = nd + 32 + 4
        self.conv2_0 = conv(od, 128, kernel_size=3, stride=1)
        self.conv2_1 = conv(od + dd[0], 128, kernel_size=3, stride=1)
        self.conv2_2 = conv(od + dd[1], 96, kernel_size=3, stride=1)
        self.conv2_3 = conv(od + dd[2], 64, kernel_size=3, stride=1)
        self.conv2_4 = conv(od + dd[3], 32, kernel_size=3, stride=1)
        self.predict_flow2 = predict_flow(od + dd[4])

        for m in self.modules():
            if isinstance(m, nn.Conv2d) or isinstance(m, nn.ConvTranspose2d):
                nn.init.kaiming_normal_(m.weight.data, mode="fan_in")
                if m.bias is not None:
                    m.bias.data.zero_()
        self.upsample1 = nn.Upsample(
            scale_factor=4, mode="bilinear", align_corners=True
        )

    def warp(self, x, flo):
        """
        warp an image/tensor (im2) back to im1, according to the optical flow
        x: [B, C, H, W] (im2)
        flo: [B, 2, H, W] flow
        """
        B, C, H, W = x.size()
        # mesh grid
        xx = torch.arange(0, W).view(1, -1).repeat(H, 1)
        yy = torch.arange(0, H).view(-1, 1).repeat(1, W)
        xx = xx.view(1, 1, H, W).repeat(B, 1, 1, 1)
        yy = yy.view(1, 1, H, W).repeat(B, 1, 1, 1)
        grid = torch.cat((xx, yy), 1).float()

        if x.is_cuda:
            grid = grid.to(dtype=x.dtype, device=x.device)
        vgrid = grid + flo

        # scale grid to [-1,1]
        vgrid[:, 0, :, :] = 2.0 * vgrid[:, 0, :, :].clone() / max(W - 1, 1) - 1.0
        vgrid[:, 1, :, :] = 2.0 * vgrid[:, 1, :, :].clone() / max(H - 1, 1) - 1.0

        vgrid = vgrid.permute(0, 2, 3, 1)
        output = nn.functional.grid_sample(x, vgrid, align_corners=True)
        mask = torch.ones(x.size()).to(dtype=x.dtype, device=x.device)
        mask = nn.functional.grid_sample(mask, vgrid, align_corners=True)

        mask[mask < 0.9999] = 0
        mask[mask > 0] = 1

        return output * mask

    def forward(self, inputs):
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=0.0,
            bgr_mult=1.0,
            bgr_to_rgb=False,
            resize_mode="interpolation",
            interpolation_mode="bilinear",
            interpolation_align_corners=False,
        )

        im1 = images[:, 0]
        im2 = images[:, 1]

        c11 = self.conv1b(self.conv1aa(self.conv1a(im1)))
        c21 = self.conv1b(self.conv1aa(self.conv1a(im2)))
        c12 = self.conv2b(self.conv2aa(self.conv2a(c11)))
        c22 = self.conv2b(self.conv2aa(self.conv2a(c21)))
        c13 = self.conv3b(self.conv3aa(self.conv3a(c12)))
        c23 = self.conv3b(self.conv3aa(self.conv3a(c22)))
        c14 = self.conv4b(self.conv4aa(self.conv4a(c13)))
        c24 = self.conv4b(self.conv4aa(self.conv4a(c23)))
        c15 = self.conv5b(self.conv5aa(self.conv5a(c14)))
        c25 = self.conv5b(self.conv5aa(self.conv5a(c24)))
        c16 = self.conv6b(self.conv6a(self.conv6aa(c15)))
        c26 = self.conv6b(self.conv6a(self.conv6aa(c25)))

        corr6 = self.corr(c16, c26)
        corr6 = corr6.view(corr6.shape[0], -1, corr6.shape[3], corr6.shape[4])
        corr6 = corr6 / c16.shape[1]
        corr6 = self.leakyRELU(corr6)

        x = torch.cat((self.conv6_0(corr6), corr6), 1)
        x = torch.cat((self.conv6_1(x), x), 1)
        x = torch.cat((self.conv6_2(x), x), 1)
        x = torch.cat((self.conv6_3(x), x), 1)
        x = torch.cat((self.conv6_4(x), x), 1)
        flow6 = self.predict_flow6(x)
        up_flow6 = self.deconv6(flow6)
        up_feat6 = self.upfeat6(x)

        warp5 = self.warp(c25, up_flow6 * 0.625)
        corr5 = self.corr(c15, warp5)
        corr5 = corr5.view(corr5.shape[0], -1, corr5.shape[3], corr5.shape[4])
        corr5 = corr5 / c15.shape[1]
        corr5 = self.leakyRELU(corr5)
        x = torch.cat((corr5, c15, up_flow6, up_feat6), 1)
        x = torch.cat((self.conv5_0(x), x), 1)
        x = torch.cat((self.conv5_1(x), x), 1)
        x = torch.cat((self.conv5_2(x), x), 1)
        x = torch.cat((self.conv5_3(x), x), 1)
        x = torch.cat((self.conv5_4(x), x), 1)
        flow5 = self.predict_flow5(x)
        up_flow5 = self.deconv5(flow5)
        up_feat5 = self.upfeat5(x)

        warp4 = self.warp(c24, up_flow5 * 1.25)
        corr4 = self.corr(c14, warp4)
        corr4 = corr4.view(corr4.shape[0], -1, corr4.shape[3], corr4.shape[4])
        corr4 = corr4 / c14.shape[1]
        corr4 = self.leakyRELU(corr4)
        x = torch.cat((corr4, c14, up_flow5, up_feat5), 1)
        x = torch.cat((self.conv4_0(x), x), 1)
        x = torch.cat((self.conv4_1(x), x), 1)
        x = torch.cat((self.conv4_2(x), x), 1)
        x = torch.cat((self.conv4_3(x), x), 1)
        x = torch.cat((self.conv4_4(x), x), 1)
        flow4 = self.predict_flow4(x)
        up_flow4 = self.deconv4(flow4)
        up_feat4 = self.upfeat4(x)

        warp3 = self.warp(c23, up_flow4 * 2.5)
        corr3 = self.corr(c13, warp3)
        corr3 = corr3.view(corr3.shape[0], -1, corr3.shape[3], corr3.shape[4])
        corr3 = corr3 / c13.shape[1]
        corr3 = self.leakyRELU(corr3)

        x = torch.cat((corr3, c13, up_flow4, up_feat4), 1)
        x = torch.cat((self.conv3_0(x), x), 1)
        x = torch.cat((self.conv3_1(x), x), 1)
        x = torch.cat((self.conv3_2(x), x), 1)
        x = torch.cat((self.conv3_3(x), x), 1)
        x = torch.cat((self.conv3_4(x), x), 1)
        flow3 = self.predict_flow3(x)
        up_flow3 = self.deconv3(flow3)
        up_feat3 = self.upfeat3(x)

        warp2 = self.warp(c22, up_flow3 * 5.0)
        corr2 = self.corr(c12, warp2)
        corr2 = corr2.view(corr2.shape[0], -1, corr2.shape[3], corr2.shape[4])
        corr2 = corr2 / c12.shape[1]
        corr2 = self.leakyRELU(corr2)
        x = torch.cat((corr2, c12, up_flow3, up_feat3), 1)
        x = torch.cat((self.conv2_0(x), x), 1)
        x = torch.cat((self.conv2_1(x), x), 1)
        x = torch.cat((self.conv2_2(x), x), 1)
        x = torch.cat((self.conv2_3(x), x), 1)
        x = torch.cat((self.conv2_4(x), x), 1)
        flow2 = self.predict_flow2(x)

        flow_up = self.upsample1(flow2 * self.div_flow)
        flow_up = self.postprocess_predictions(flow_up, image_resizer, is_flow=True)

        outputs = {}
        if self.training:
            outputs["flow_preds"] = [flow2, flow3, flow4, flow5, flow6]
            outputs["flows"] = flow_up[:, None]
        else:
            outputs["flows"] = flow_up[:, None]
        return outputs


class PWCDCNet(PWCNet):
    pretrained_checkpoints = {
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/pwcdcnet-things-cc223701.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/pwcdcnet-sintel-c7d08a46.ckpt",
    }

    def __init__(
        self,
        div_flow: float = 20.0,
        md: int = 4,
        loss_start_scale: int = 4,
        loss_num_scales: int = 5,
        loss_base_weight: float = 0.32,
        loss_norm: str = "L2",
        **kwargs,
    ):
        super(PWCDCNet, self).__init__(
            div_flow=div_flow,
            md=md,
            loss_start_scale=loss_start_scale,
            loss_num_scales=loss_num_scales,
            loss_base_weight=loss_base_weight,
            loss_norm=loss_norm,
            **kwargs,
        )

        nd = (2 * self.md + 1) ** 2
        dd = np.cumsum([128, 128, 96, 64, 32])
        od = nd + 32 + 4

        self.dc_conv1 = conv(
            od + dd[4], 128, kernel_size=3, stride=1, padding=1, dilation=1
        )
        self.dc_conv2 = conv(128, 128, kernel_size=3, stride=1, padding=2, dilation=2)
        self.dc_conv3 = conv(128, 128, kernel_size=3, stride=1, padding=4, dilation=4)
        self.dc_conv4 = conv(128, 96, kernel_size=3, stride=1, padding=8, dilation=8)
        self.dc_conv5 = conv(96, 64, kernel_size=3, stride=1, padding=16, dilation=16)
        self.dc_conv6 = conv(64, 32, kernel_size=3, stride=1, padding=1, dilation=1)
        self.dc_conv7 = predict_flow(32)

        for m in self.modules():
            if isinstance(m, nn.Conv2d) or isinstance(m, nn.ConvTranspose2d):
                nn.init.kaiming_normal_(m.weight.data, mode="fan_in")
                if m.bias is not None:
                    m.bias.data.zero_()

    def forward(self, inputs):
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=0.0,
            bgr_mult=1.0,
            bgr_to_rgb=True,
            resize_mode="interpolation",
            interpolation_mode="bilinear",
            interpolation_align_corners=True,
        )

        im1 = images[:, 0]
        im2 = images[:, 1]

        c11 = self.conv1b(self.conv1aa(self.conv1a(im1)))
        c21 = self.conv1b(self.conv1aa(self.conv1a(im2)))
        c12 = self.conv2b(self.conv2aa(self.conv2a(c11)))
        c22 = self.conv2b(self.conv2aa(self.conv2a(c21)))
        c13 = self.conv3b(self.conv3aa(self.conv3a(c12)))
        c23 = self.conv3b(self.conv3aa(self.conv3a(c22)))
        c14 = self.conv4b(self.conv4aa(self.conv4a(c13)))
        c24 = self.conv4b(self.conv4aa(self.conv4a(c23)))
        c15 = self.conv5b(self.conv5aa(self.conv5a(c14)))
        c25 = self.conv5b(self.conv5aa(self.conv5a(c24)))
        c16 = self.conv6b(self.conv6a(self.conv6aa(c15)))
        c26 = self.conv6b(self.conv6a(self.conv6aa(c25)))

        corr6 = self.corr(c16, c26)
        corr6 = corr6.view(corr6.shape[0], -1, corr6.shape[3], corr6.shape[4])
        corr6 = corr6 / c16.shape[1]
        corr6 = self.leakyRELU(corr6)

        x = torch.cat((self.conv6_0(corr6), corr6), 1)
        x = torch.cat((self.conv6_1(x), x), 1)
        x = torch.cat((self.conv6_2(x), x), 1)
        x = torch.cat((self.conv6_3(x), x), 1)
        x = torch.cat((self.conv6_4(x), x), 1)
        flow6 = self.predict_flow6(x)
        up_flow6 = self.deconv6(flow6)
        up_feat6 = self.upfeat6(x)

        warp5 = self.warp(c25, up_flow6 * 0.625)
        corr5 = self.corr(c15, warp5)
        corr5 = corr5.view(corr5.shape[0], -1, corr5.shape[3], corr5.shape[4])
        corr5 = corr5 / c15.shape[1]
        corr5 = self.leakyRELU(corr5)
        x = torch.cat((corr5, c15, up_flow6, up_feat6), 1)
        x = torch.cat((self.conv5_0(x), x), 1)
        x = torch.cat((self.conv5_1(x), x), 1)
        x = torch.cat((self.conv5_2(x), x), 1)
        x = torch.cat((self.conv5_3(x), x), 1)
        x = torch.cat((self.conv5_4(x), x), 1)
        flow5 = self.predict_flow5(x)
        up_flow5 = self.deconv5(flow5)
        up_feat5 = self.upfeat5(x)

        warp4 = self.warp(c24, up_flow5 * 1.25)
        corr4 = self.corr(c14, warp4)
        corr4 = corr4.view(corr4.shape[0], -1, corr4.shape[3], corr4.shape[4])
        corr4 = corr4 / c14.shape[1]
        corr4 = self.leakyRELU(corr4)
        x = torch.cat((corr4, c14, up_flow5, up_feat5), 1)
        x = torch.cat((self.conv4_0(x), x), 1)
        x = torch.cat((self.conv4_1(x), x), 1)
        x = torch.cat((self.conv4_2(x), x), 1)
        x = torch.cat((self.conv4_3(x), x), 1)
        x = torch.cat((self.conv4_4(x), x), 1)
        flow4 = self.predict_flow4(x)
        up_flow4 = self.deconv4(flow4)
        up_feat4 = self.upfeat4(x)

        warp3 = self.warp(c23, up_flow4 * 2.5)
        corr3 = self.corr(c13, warp3)
        corr3 = corr3.view(corr3.shape[0], -1, corr3.shape[3], corr3.shape[4])
        corr3 = corr3 / c13.shape[1]
        corr3 = self.leakyRELU(corr3)

        x = torch.cat((corr3, c13, up_flow4, up_feat4), 1)
        x = torch.cat((self.conv3_0(x), x), 1)
        x = torch.cat((self.conv3_1(x), x), 1)
        x = torch.cat((self.conv3_2(x), x), 1)
        x = torch.cat((self.conv3_3(x), x), 1)
        x = torch.cat((self.conv3_4(x), x), 1)
        flow3 = self.predict_flow3(x)
        up_flow3 = self.deconv3(flow3)
        up_feat3 = self.upfeat3(x)

        warp2 = self.warp(c22, up_flow3 * 5.0)
        corr2 = self.corr(c12, warp2)
        corr2 = corr2.view(corr2.shape[0], -1, corr2.shape[3], corr2.shape[4])
        corr2 = corr2 / c12.shape[1]
        corr2 = self.leakyRELU(corr2)
        x = torch.cat((corr2, c12, up_flow3, up_feat3), 1)
        x = torch.cat((self.conv2_0(x), x), 1)
        x = torch.cat((self.conv2_1(x), x), 1)
        x = torch.cat((self.conv2_2(x), x), 1)
        x = torch.cat((self.conv2_3(x), x), 1)
        x = torch.cat((self.conv2_4(x), x), 1)
        flow2 = self.predict_flow2(x)

        x = self.dc_conv4(self.dc_conv3(self.dc_conv2(self.dc_conv1(x))))
        flow2 = flow2 + self.dc_conv7(self.dc_conv6(self.dc_conv5(x)))

        flow_up = self.upsample1(flow2 * self.div_flow)
        flow_up = self.postprocess_predictions(flow_up, image_resizer, is_flow=True)

        outputs = {}
        if self.training:
            outputs["flow_preds"] = [flow2, flow3, flow4, flow5, flow6]
            outputs["flows"] = flow_up[:, None]
        else:
            outputs["flows"] = flow_up[:, None]
        return outputs


@register_model
@trainable
class pwcnet(PWCDCNet):
    pass


@register_model
@trainable
class pwcnet_nodc(PWCNet):
    pass



from typing import Optional, Sequence

from einops import rearrange
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
import numpy as np
from torchvision import ops

try:
    from spatial_correlation_sampler import SpatialCorrelationSampler
except ModuleNotFoundError:
    from ptlflow.utils.correlation import (
        IterSpatialCorrelationSampler as SpatialCorrelationSampler,
    )

from ptlflow.utils.registry import register_model, trainable
from ..base_model.base_model import BaseModel


def conv(
    in_planes,
    out_planes,
    kernel_size=3,
    stride=1,
    padding=1,
    dilation=1,
    activation=True,
):
    if activation:
        return nn.Sequential(
            nn.Conv2d(
                in_planes,
                out_planes,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                dilation=dilation,
                bias=True,
            ),
            nn.LeakyReLU(0.1),
        )
    else:
        return nn.Sequential(
            nn.Conv2d(
                in_planes,
                out_planes,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                dilation=dilation,
                bias=True,
            )
        )


def predict_flow(in_planes):
    return nn.Conv2d(in_planes, 2, kernel_size=3, stride=1, padding=1, bias=True)


def predict_mask(in_planes):
    return nn.Conv2d(in_planes, 1, kernel_size=3, stride=1, padding=1, bias=True)


def deconv(in_planes, out_planes, kernel_size=4, stride=2, padding=1):
    return nn.ConvTranspose2d(
        in_planes, out_planes, kernel_size, stride, padding, bias=True
    )


def deformable_conv(
    in_planes, out_planes, kernel_size=3, strides=1, padding=1, use_bias=True
):
    return ops.DeformConv2d(
        in_planes, out_planes, kernel_size, strides, padding, bias=use_bias
    )


def upsample_kernel2d(w, dtype, device):
    c = w // 2
    kernel = 1 - torch.abs(c - torch.arange(w, dtype=dtype, device=device)) / (c + 1)
    kernel = kernel.repeat(w).view(w, -1) * kernel.unsqueeze(1)
    return kernel.view(1, 1, w, w)


def downsample_kernel2d(w, device):
    kernel = (
        (w + 1) - torch.abs(w - torch.arange(w * 2 + 1, dtype=dtype, device=device))
    ) / (2 * w + 1)
    kernel = kernel.repeat(w).view(w, -1) * kernel.unsqueeze(1)
    return kernel.view(1, 1, w * 2 + 1, w * 2 + 1)


def Upsample(img, factor):
    if factor == 1:
        return img
    B, C, H, W = img.shape
    batch_img = img.view(B * C, 1, H, W)
    batch_img = F.pad(batch_img, [0, 1, 0, 1], mode="replicate")
    kernel = upsample_kernel2d(factor * 2 - 1, img.dtype, img.device)
    upsamp_img = F.conv_transpose2d(
        batch_img, kernel, stride=factor, padding=(factor - 1)
    )
    upsamp_img = upsamp_img[:, :, :-1, :-1]
    _, _, H_up, W_up = upsamp_img.shape
    return upsamp_img.view(B, C, H_up, W_up)


def Downsample(img, factor):
    if factor == 1:
        return img
    B, C, H, W = img.shape
    batch_img = img.view(B * C, 1, H, W)
    kernel = downsample_kernel2d(factor // 2, img.dtype, img.device)
    upsamp_img = F.conv2d(batch_img, kernel, stride=factor, padding=factor // 2)
    upsamp_nom = F.conv2d(
        torch.ones_like(batch_img), kernel, stride=factor, padding=factor // 2
    )
    _, _, H_up, W_up = upsamp_img.shape
    upsamp_img = upsamp_img.view(B, C, H_up, W_up)
    upsamp_nom = upsamp_nom.view(B, C, H_up, W_up)
    return upsamp_img / upsamp_nom


class MaskFlownet_S(BaseModel):
    pretrained_checkpoints = {
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/maskflownet_s-sintel-79196ff3.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/maskflownet_s-things-a8a8ed9e.ckpt",
    }

    """
    PWC-DC net. add dilation convolution and densenet connections
    """

    def __init__(
        self,
        div_flow: float = 20.0,
        md: int = 4,
        flow_multiplier: float = 1.0,
        deform_bias: bool = True,
        strides: Sequence[int] = (64, 32, 16, 8, 4),
        upfeat_ch: Sequence[int] = (16, 16, 16, 16),
        loss_weights: Sequence[float] = (0.005, 0.01, 0.02, 0.08, 0.32),
        loss_match: str = "upsampling",
        loss_eps: float = 1e-8,
        loss_q: Optional[float] = None,
        **kwargs,
    ):
        super(MaskFlownet_S, self).__init__(
            loss_fn=MultiscaleEpe(
                scales=strides,
                weights=loss_weights,
                match=loss_match,
                eps=loss_eps,
                q=loss_q,
            ),
            output_stride=64,
            **kwargs,
        )

        self.div_flow = div_flow
        self.md = md
        self.flow_multiplier = flow_multiplier
        self.deform_bias = deform_bias
        self.strides = strides
        self.upfeat_ch = upfeat_ch

        self.scale = self.div_flow * self.flow_multiplier
        self.deform_bias = self.deform_bias

        self.conv1a = conv(3, 16, kernel_size=3, stride=2)
        self.conv1b = conv(16, 16, kernel_size=3, stride=1)
        self.conv1c = conv(16, 16, kernel_size=3, stride=1)
        self.conv2a = conv(16, 32, kernel_size=3, stride=2)
        self.conv2b = conv(32, 32, kernel_size=3, stride=1)
        self.conv2c = conv(32, 32, kernel_size=3, stride=1)
        self.conv3a = conv(32, 64, kernel_size=3, stride=2)
        self.conv3b = conv(64, 64, kernel_size=3, stride=1)
        self.conv3c = conv(64, 64, kernel_size=3, stride=1)
        self.conv4a = conv(64, 96, kernel_size=3, stride=2)
        self.conv4b = conv(96, 96, kernel_size=3, stride=1)
        self.conv4c = conv(96, 96, kernel_size=3, stride=1)
        self.conv5a = conv(96, 128, kernel_size=3, stride=2)
        self.conv5b = conv(128, 128, kernel_size=3, stride=1)
        self.conv5c = conv(128, 128, kernel_size=3, stride=1)
        self.conv6a = conv(128, 196, kernel_size=3, stride=2)
        self.conv6b = conv(196, 196, kernel_size=3, stride=1)
        self.conv6c = conv(196, 196, kernel_size=3, stride=1)

        self.corr_fn = SpatialCorrelationSampler(
            kernel_size=1, patch_size=2 * self.md + 1, padding=0
        )
        # self.corr    = Correlation(pad_size=md, kernel_size=1, max_displacement=md, stride1=1, stride2=1, corr_multiply=1)
        self.leakyRELU = nn.LeakyReLU(0.1)

        nd = (2 * self.md + 1) ** 2
        dd = np.cumsum([128, 128, 96, 64, 32])

        od = nd
        self.conv6_0 = conv(od, 128, kernel_size=3, stride=1)
        self.conv6_1 = conv(od + dd[0], 128, kernel_size=3, stride=1)
        self.conv6_2 = conv(od + dd[1], 96, kernel_size=3, stride=1)
        self.conv6_3 = conv(od + dd[2], 64, kernel_size=3, stride=1)
        self.conv6_4 = conv(od + dd[3], 32, kernel_size=3, stride=1)
        self.pred_flow6 = predict_flow(od + dd[4])
        self.pred_mask6 = predict_mask(od + dd[4])
        self.upfeat5 = deconv(
            od + dd[4], self.upfeat_ch[0], kernel_size=4, stride=2, padding=1
        )
        # self.deconv6 = deconv(2, 2, kernel_size=4, stride=2, padding=1)
        # self.upfeat6 = deconv(od+dd[4], 2, kernel_size=4, stride=2, padding=1)

        # od = nd+128+4
        od = nd + 128 + 18
        self.conv5_0 = conv(od, 128, kernel_size=3, stride=1)
        self.conv5_1 = conv(od + dd[0], 128, kernel_size=3, stride=1)
        self.conv5_2 = conv(od + dd[1], 96, kernel_size=3, stride=1)
        self.conv5_3 = conv(od + dd[2], 64, kernel_size=3, stride=1)
        self.conv5_4 = conv(od + dd[3], 32, kernel_size=3, stride=1)
        self.pred_flow5 = predict_flow(od + dd[4])
        self.pred_mask5 = predict_mask(od + dd[4])
        self.upfeat4 = deconv(
            od + dd[4], self.upfeat_ch[1], kernel_size=4, stride=2, padding=1
        )
        # self.deconv5 = deconv(2, 2, kernel_size=4, stride=2, padding=1)
        # self.upfeat5 = deconv(od+dd[4], 2, kernel_size=4, stride=2, padding=1)

        # od = nd+96+4
        od = nd + 96 + 18
        self.conv4_0 = conv(od, 128, kernel_size=3, stride=1)
        self.conv4_1 = conv(od + dd[0], 128, kernel_size=3, stride=1)
        self.conv4_2 = conv(od + dd[1], 96, kernel_size=3, stride=1)
        self.conv4_3 = conv(od + dd[2], 64, kernel_size=3, stride=1)
        self.conv4_4 = conv(od + dd[3], 32, kernel_size=3, stride=1)
        self.pred_flow4 = predict_flow(od + dd[4])
        self.pred_mask4 = predict_mask(od + dd[4])
        self.upfeat3 = deconv(
            od + dd[4], self.upfeat_ch[2], kernel_size=4, stride=2, padding=1
        )
        # self.deconv4 = deconv(2, 2, kernel_size=4, stride=2, padding=1)
        # self.upfeat4 = deconv(od+dd[4], 2, kernel_size=4, stride=2, padding=1)

        # od = nd+64+4
        od = nd + 64 + 18
        self.conv3_0 = conv(od, 128, kernel_size=3, stride=1)
        self.conv3_1 = conv(od + dd[0], 128, kernel_size=3, stride=1)
        self.conv3_2 = conv(od + dd[1], 96, kernel_size=3, stride=1)
        self.conv3_3 = conv(od + dd[2], 64, kernel_size=3, stride=1)
        self.conv3_4 = conv(od + dd[3], 32, kernel_size=3, stride=1)
        self.pred_flow3 = predict_flow(od + dd[4])
        self.pred_mask3 = predict_mask(od + dd[4])
        self.upfeat2 = deconv(
            od + dd[4], self.upfeat_ch[3], kernel_size=4, stride=2, padding=1
        )
        # self.deconv3 = deconv(2, 2, kernel_size=4, stride=2, padding=1)
        # self.upfeat3 = deconv(od+dd[4], 2, kernel_size=4, stride=2, padding=1)

        # od = nd+32+4
        od = nd + 32 + 18
        self.conv2_0 = conv(od, 128, kernel_size=3, stride=1)
        self.conv2_1 = conv(od + dd[0], 128, kernel_size=3, stride=1)
        self.conv2_2 = conv(od + dd[1], 96, kernel_size=3, stride=1)
        self.conv2_3 = conv(od + dd[2], 64, kernel_size=3, stride=1)
        self.conv2_4 = conv(od + dd[3], 32, kernel_size=3, stride=1)
        self.pred_flow2 = predict_flow(od + dd[4])

        self.dc_conv1 = conv(
            od + dd[4], 128, kernel_size=3, stride=1, padding=1, dilation=1
        )
        self.dc_conv2 = conv(128, 128, kernel_size=3, stride=1, padding=2, dilation=2)
        self.dc_conv3 = conv(128, 128, kernel_size=3, stride=1, padding=4, dilation=4)
        self.dc_conv4 = conv(128, 96, kernel_size=3, stride=1, padding=8, dilation=8)
        self.dc_conv5 = conv(96, 64, kernel_size=3, stride=1, padding=16, dilation=16)
        self.dc_conv6 = conv(64, 32, kernel_size=3, stride=1, padding=1, dilation=1)
        self.dc_conv7 = predict_flow(32)

        # self.upfeat5 = deconv()

        self.deform5 = deformable_conv(128, 128)
        self.deform4 = deformable_conv(96, 96)
        self.deform3 = deformable_conv(64, 64)
        self.deform2 = deformable_conv(32, 32)

        self.conv5f = conv(
            16, 128, kernel_size=3, stride=1, padding=1, activation=False
        )
        self.conv4f = conv(16, 96, kernel_size=3, stride=1, padding=1, activation=False)
        self.conv3f = conv(16, 64, kernel_size=3, stride=1, padding=1, activation=False)
        self.conv2f = conv(16, 32, kernel_size=3, stride=1, padding=1, activation=False)

        for m in self.modules():
            if isinstance(m, nn.Conv2d) or isinstance(m, nn.ConvTranspose2d):
                nn.init.kaiming_normal_(m.weight.data, mode="fan_in")
                if m.bias is not None:
                    m.bias.data.zero_()

    def corr(self, f1, f2):
        corr = self.corr_fn(f1, f2)
        corr = corr.view(corr.shape[0], -1, corr.shape[3], corr.shape[4])
        corr = corr / f2.shape[1]
        return corr

    def warp(self, x, flo):
        """
        warp an image/tensor (im2) back to im1, according to the optical flow
        x: [B, C, H, W] (im2)
        flo: [B, 2, H, W] flow
        """
        B, C, H, W = x.size()
        # mesh grid
        xx = torch.arange(0, W, dtype=x.dtype, device=x.device).view(1, -1).repeat(H, 1)
        yy = torch.arange(0, H, dtype=x.dtype, device=x.device).view(-1, 1).repeat(1, W)
        xx = xx.view(1, 1, H, W).repeat(B, 1, 1, 1)
        yy = yy.view(1, 1, H, W).repeat(B, 1, 1, 1)
        grid = torch.cat((xx, yy), 1)

        device = x.device
        grid = grid.to(device)
        # vgrid = Variable(grid) + flo
        vgrid = Variable(grid) + torch.flip(flo, [1])

        # scale grid to [-1,1]
        vgrid[:, 0, :, :] = 2.0 * vgrid[:, 0, :, :].clone() / max(W - 1, 1) - 1.0
        vgrid[:, 1, :, :] = 2.0 * vgrid[:, 1, :, :].clone() / max(H - 1, 1) - 1.0

        vgrid = vgrid.permute(0, 2, 3, 1)
        # vgrid = vgrid.permute(0,2,3,1).clamp(-1.1, 1.1)
        output = nn.functional.grid_sample(x, vgrid, align_corners=True)
        mask = torch.autograd.Variable(
            torch.ones(x.size(), dtype=x.dtype, device=device)
        )
        mask = nn.functional.grid_sample(mask, vgrid, align_corners=True)

        # if W==128:
        # np.save('mask.npy', mask.cpu().data.numpy())
        # np.save('warp.npy', output.cpu().data.numpy())

        mask[mask < 0.9999] = 0
        mask[mask > 0] = 1

        return output * mask

    def forward(self, inputs, skip_preprocess=False):
        if not skip_preprocess:
            bgr_mean = rearrange(inputs["images"], "b n c h w -> b c (n h w)").mean(2)[
                :, None, :, None, None
            ]
            images, image_resizer = self.preprocess_images(
                inputs["images"],
                bgr_add=-bgr_mean,
                bgr_mult=1.0,
                bgr_to_rgb=True,
                resize_mode="interpolation",
                interpolation_mode="bilinear",
                interpolation_align_corners=False,
            )

        im1 = images[:, 0]
        im2 = images[:, 1]

        c11 = self.conv1c(self.conv1b(self.conv1a(im1)))
        c21 = self.conv1c(self.conv1b(self.conv1a(im2)))
        c12 = self.conv2c(self.conv2b(self.conv2a(c11)))
        c22 = self.conv2c(self.conv2b(self.conv2a(c21)))
        c13 = self.conv3c(self.conv3b(self.conv3a(c12)))
        c23 = self.conv3c(self.conv3b(self.conv3a(c22)))
        c14 = self.conv4c(self.conv4b(self.conv4a(c13)))
        c24 = self.conv4c(self.conv4b(self.conv4a(c23)))
        c15 = self.conv5c(self.conv5b(self.conv5a(c14)))
        c25 = self.conv5c(self.conv5b(self.conv5a(c24)))
        c16 = self.conv6c(self.conv6b(self.conv6a(c15)))
        c26 = self.conv6c(self.conv6b(self.conv6a(c25)))

        corr6 = self.corr(c16, c26)
        corr6 = self.leakyRELU(corr6)

        x = torch.cat((self.conv6_0(corr6), corr6), 1)
        x = torch.cat((self.conv6_1(x), x), 1)
        x = torch.cat((self.conv6_2(x), x), 1)
        x = torch.cat((self.conv6_3(x), x), 1)
        x = torch.cat((self.conv6_4(x), x), 1)
        flow6 = self.pred_flow6(x)
        mask6 = self.pred_mask6(x)

        feat5 = self.leakyRELU(self.upfeat5(x))
        flow5 = Upsample(flow6, 2)
        mask5 = Upsample(mask6, 2)
        warp5 = (flow5 * self.scale / self.strides[1]).unsqueeze(1)
        warp5 = torch.repeat_interleave(warp5, 9, 1)
        S1, S2, S3, S4, S5 = warp5.shape
        warp5 = warp5.view(S1, S2 * S3, S4, S5)
        warp5 = self.deform5(c25, warp5)
        tradeoff5 = feat5
        warp5 = (warp5 * torch.sigmoid(mask5)) + self.conv5f(tradeoff5)
        warp5 = self.leakyRELU(warp5)
        corr5 = self.corr(c15, warp5)
        corr5 = self.leakyRELU(corr5)
        x = torch.cat((corr5, c15, feat5, flow5), 1)
        x = torch.cat((self.conv5_0(x), x), 1)
        x = torch.cat((self.conv5_1(x), x), 1)
        x = torch.cat((self.conv5_2(x), x), 1)
        x = torch.cat((self.conv5_3(x), x), 1)
        x = torch.cat((self.conv5_4(x), x), 1)
        flow5 = flow5 + self.pred_flow5(x)
        mask5 = self.pred_mask5(x)

        feat4 = self.leakyRELU(self.upfeat4(x))
        flow4 = Upsample(flow5, 2)
        mask4 = Upsample(mask5, 2)
        warp4 = (flow4 * self.scale / self.strides[2]).unsqueeze(1)
        warp4 = torch.repeat_interleave(warp4, 9, 1)
        S1, S2, S3, S4, S5 = warp4.shape
        warp4 = warp4.view(S1, S2 * S3, S4, S5)
        warp4 = self.deform4(c24, warp4)
        tradeoff4 = feat4
        warp4 = (warp4 * torch.sigmoid(mask4)) + self.conv4f(tradeoff4)
        warp4 = self.leakyRELU(warp4)
        corr4 = self.corr(c14, warp4)
        corr4 = self.leakyRELU(corr4)
        x = torch.cat((corr4, c14, feat4, flow4), 1)
        x = torch.cat((self.conv4_0(x), x), 1)
        x = torch.cat((self.conv4_1(x), x), 1)
        x = torch.cat((self.conv4_2(x), x), 1)
        x = torch.cat((self.conv4_3(x), x), 1)
        x = torch.cat((self.conv4_4(x), x), 1)
        flow4 = flow4 + self.pred_flow4(x)
        mask4 = self.pred_mask4(x)

        feat3 = self.leakyRELU(self.upfeat3(x))
        flow3 = Upsample(flow4, 2)
        mask3 = Upsample(mask4, 2)
        warp3 = (flow3 * self.scale / self.strides[3]).unsqueeze(1)
        warp3 = torch.repeat_interleave(warp3, 9, 1)
        S1, S2, S3, S4, S5 = warp3.shape
        warp3 = warp3.view(S1, S2 * S3, S4, S5)
        warp3 = self.deform3(c23, warp3)
        tradeoff3 = feat3
        warp3 = (warp3 * torch.sigmoid(mask3)) + self.conv3f(tradeoff3)
        warp3 = self.leakyRELU(warp3)
        corr3 = self.corr(c13, warp3)
        corr3 = self.leakyRELU(corr3)
        x = torch.cat((corr3, c13, feat3, flow3), 1)
        x = torch.cat((self.conv3_0(x), x), 1)
        x = torch.cat((self.conv3_1(x), x), 1)
        x = torch.cat((self.conv3_2(x), x), 1)
        x = torch.cat((self.conv3_3(x), x), 1)
        x = torch.cat((self.conv3_4(x), x), 1)
        flow3 = flow3 + self.pred_flow3(x)
        mask3 = self.pred_mask3(x)

        feat2 = self.leakyRELU(self.upfeat2(x))
        flow2 = Upsample(flow3, 2)
        mask2 = Upsample(mask3, 2)
        warp2 = (flow2 * self.scale / self.strides[4]).unsqueeze(1)
        warp2 = torch.repeat_interleave(warp2, 9, 1)
        S1, S2, S3, S4, S5 = warp2.shape
        warp2 = warp2.view(S1, S2 * S3, S4, S5)
        warp2 = self.deform2(c22, warp2)
        tradeoff2 = feat2
        warp2 = (warp2 * torch.sigmoid(mask2)) + self.conv2f(tradeoff2)
        warp2 = self.leakyRELU(warp2)
        corr2 = self.corr(c12, warp2)
        corr2 = self.leakyRELU(corr2)
        x = torch.cat((corr2, c12, feat2, flow2), 1)
        x = torch.cat((self.conv2_0(x), x), 1)
        x = torch.cat((self.conv2_1(x), x), 1)
        x = torch.cat((self.conv2_2(x), x), 1)
        x = torch.cat((self.conv2_3(x), x), 1)
        x = torch.cat((self.conv2_4(x), x), 1)
        flow2 = flow2 + self.pred_flow2(x)

        x = self.dc_conv4(self.dc_conv3(self.dc_conv2(self.dc_conv1(x))))
        flow2 = flow2 + self.dc_conv7(self.dc_conv6(self.dc_conv5(x)))

        predictions = [
            flow.flip(1) * self.scale for flow in [flow6, flow5, flow4, flow3, flow2]
        ]
        occlusion_masks = []
        occlusion_masks.append(1 - torch.sigmoid(mask2))
        c1s = [c11, c12, c13, c14, c15, c16]
        c2s = [c21, c12, c13, c24, c25, c26]
        flows = [flow6, flow5, flow4, flow3, flow2]
        mask0 = Upsample(mask2, 4)
        mask0 = torch.sigmoid(mask0) - 0.5
        c30 = im1
        c40 = self.warp(im2, Upsample(flow2, 4) * self.scale)
        c30 = torch.cat((c30, torch.zeros_like(mask0)), 1)
        c40 = torch.cat((c40, mask0), 1)
        srcs = [c1s, c2s, flows, c30, c40]

        flow_up = F.interpolate(
            predictions[-1],
            size=im1.shape[-2:],
            mode="bilinear",
            align_corners=True,
        )
        occ_up = F.interpolate(
            occlusion_masks[-1],
            size=im1.shape[-2:],
            mode="bilinear",
            align_corners=True,
        )
        if not skip_preprocess:
            flow_up = self.postprocess_predictions(flow_up, image_resizer, is_flow=True)
            occ_up = self.postprocess_predictions(occ_up, image_resizer, is_flow=False)

        output = {
            "flows": flow_up[:, None],
            "occs": occ_up[:, None],
            "srcs": srcs,
        }
        if self.training:
            output["flow_preds"] = predictions
            output["occ_preds"] = occlusion_masks

        return output


class MaskFlownet(BaseModel):
    pretrained_checkpoints = {
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/maskflownet-kitti-d2ccb09b.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/maskflownet-sintel-c52423aa.ckpt",
    }

    def __init__(
        self,
        div_flow: float = 20.0,
        md: int = 4,
        md2: int = 2,
        flow_multiplier: float = 1.0,
        deform_bias: bool = True,
        strides: Sequence[int] = (64, 32, 16, 8, 4),
        upfeat_ch: Sequence[int] = (16, 16, 16, 16),
        loss_weights: Sequence[float] = (0.005, 0.01, 0.02, 0.08, 0.32),
        loss_match: str = "upsampling",
        loss_eps: float = 1e-8,
        loss_q: Optional[float] = None,
        **kwargs,
    ):
        super(MaskFlownet, self).__init__(
            loss_fn=MultiscaleEpe(
                scales=strides,
                weights=loss_weights,
                match=loss_match,
                eps=loss_eps,
                q=loss_q,
            ),
            output_stride=64,
            **kwargs,
        )

        self.div_flow = div_flow
        self.md = md
        self.md2 = md2
        self.flow_multiplier = flow_multiplier
        self.deform_bias = deform_bias
        self.strides = strides
        self.upfeat_ch = upfeat_ch

        self.scale = self.div_flow * self.flow_multiplier
        self.deform_bias = self.deform_bias

        self.MaskFlownet_S = MaskFlownet_S(
            div_flow=div_flow,
            md=md,
            flow_multiplier=flow_multiplier,
            deform_bias=deform_bias,
            strides=strides,
            upfeat_ch=upfeat_ch,
            loss_weights=loss_weights,
            loss_match=loss_match,
            loss_eps=loss_eps,
            loss_q=loss_q,
            **kwargs,
        )
        self.activate = nn.LeakyReLU(0.1)

        self.conv1x = conv(4, 16, stride=2)
        self.conv1y = conv(16, 16, stride=1)
        self.conv1z = conv(16, 16, stride=1)
        self.conv2x = conv(16, 32, stride=2)
        self.conv2y = conv(32, 32, stride=1)
        self.conv2z = conv(32, 32, stride=1)
        self.conv3x = conv(32, 64, stride=2)
        self.conv3y = conv(64, 64, stride=1)
        self.conv3z = conv(64, 64, stride=1)
        self.conv4x = conv(64, 96, stride=2)
        self.conv4y = conv(96, 96, stride=1)
        self.conv4z = conv(96, 96, stride=1)
        self.conv5x = conv(96, 128, stride=2)
        self.conv5y = conv(128, 128, stride=1)
        self.conv5z = conv(128, 128, stride=1)
        self.conv6x = conv(128, 196, stride=2)
        self.conv6y = conv(196, 196, stride=1)
        self.conv6z = conv(196, 196, stride=1)

        self.leakyRELU = nn.LeakyReLU(0.1)
        self.corr_fn = SpatialCorrelationSampler(
            kernel_size=1, patch_size=2 * self.md2 + 1, padding=0
        )
        # self.corr    = Correlation(pad_size=self.md, kernel_size=1, max_displacement=self.md, stride1=1, stride2=1, corr_multiply=1)

        nd = (2 * self.md2 + 1) ** 2
        dd = np.cumsum([128, 128, 96, 64, 32])

        od = nd + nd + 2
        self.conv6_0 = conv(od, 128, kernel_size=3, stride=1)
        self.conv6_1 = conv(od + dd[0], 128, kernel_size=3, stride=1)
        self.conv6_2 = conv(od + dd[1], 96, kernel_size=3, stride=1)
        self.conv6_3 = conv(od + dd[2], 64, kernel_size=3, stride=1)
        self.conv6_4 = conv(od + dd[3], 32, kernel_size=3, stride=1)
        self.pred_flow6 = predict_flow(od + dd[4])
        self.upfeat5 = deconv(
            od + dd[4], self.upfeat_ch[0], kernel_size=4, stride=2, padding=1
        )

        # od = nd+128+4
        od = nd + nd + 128 + 16 + 2 + 2
        self.conv5_0 = conv(od, 128, kernel_size=3, stride=1)
        self.conv5_1 = conv(od + dd[0], 128, kernel_size=3, stride=1)
        self.conv5_2 = conv(od + dd[1], 96, kernel_size=3, stride=1)
        self.conv5_3 = conv(od + dd[2], 64, kernel_size=3, stride=1)
        self.conv5_4 = conv(od + dd[3], 32, kernel_size=3, stride=1)
        self.pred_flow5 = predict_flow(od + dd[4])
        self.upfeat4 = deconv(
            od + dd[4], self.upfeat_ch[1], kernel_size=4, stride=2, padding=1
        )
        # self.deconv5 = deconv(2, 2, kernel_size=4, stride=2, padding=1)
        # self.upfeat5 = deconv(od+dd[4], 2, kernel_size=4, stride=2, padding=1)

        # od = nd+96+4
        # od = nd+96+18
        od = nd + nd + 96 + 16 + 2 + 2
        self.conv4_0 = conv(od, 128, kernel_size=3, stride=1)
        self.conv4_1 = conv(od + dd[0], 128, kernel_size=3, stride=1)
        self.conv4_2 = conv(od + dd[1], 96, kernel_size=3, stride=1)
        self.conv4_3 = conv(od + dd[2], 64, kernel_size=3, stride=1)
        self.conv4_4 = conv(od + dd[3], 32, kernel_size=3, stride=1)
        self.pred_flow4 = predict_flow(od + dd[4])
        self.upfeat3 = deconv(
            od + dd[4], self.upfeat_ch[2], kernel_size=4, stride=2, padding=1
        )
        # self.deconv4 = deconv(2, 2, kernel_size=4, stride=2, padding=1)
        # self.upfeat4 = deconv(od+dd[4], 2, kernel_size=4, stride=2, padding=1)

        # od = nd+64+4
        # od = nd+64+18
        od = nd + nd + 64 + 16 + 2 + 2
        self.conv3_0 = conv(od, 128, kernel_size=3, stride=1)
        self.conv3_1 = conv(od + dd[0], 128, kernel_size=3, stride=1)
        self.conv3_2 = conv(od + dd[1], 96, kernel_size=3, stride=1)
        self.conv3_3 = conv(od + dd[2], 64, kernel_size=3, stride=1)
        self.conv3_4 = conv(od + dd[3], 32, kernel_size=3, stride=1)
        self.pred_flow3 = predict_flow(od + dd[4])
        self.upfeat2 = deconv(
            od + dd[4], self.upfeat_ch[3], kernel_size=4, stride=2, padding=1
        )
        # self.deconv3 = deconv(2, 2, kernel_size=4, stride=2, padding=1)
        # self.upfeat3 = deconv(od+dd[4], 2, kernel_size=4, stride=2, padding=1)

        # od = nd+32+4
        # od = nd+32+18
        od = nd + nd + 32 + 16 + 2 + 2
        self.conv2_0 = conv(od, 128, kernel_size=3, stride=1)
        self.conv2_1 = conv(od + dd[0], 128, kernel_size=3, stride=1)
        self.conv2_2 = conv(od + dd[1], 96, kernel_size=3, stride=1)
        self.conv2_3 = conv(od + dd[2], 64, kernel_size=3, stride=1)
        self.conv2_4 = conv(od + dd[3], 32, kernel_size=3, stride=1)
        self.pred_flow2 = predict_flow(od + dd[4])

        self.dc_conv1 = conv(
            od + dd[4], 128, kernel_size=3, stride=1, padding=1, dilation=1
        )
        self.dc_conv2 = conv(128, 128, kernel_size=3, stride=1, padding=2, dilation=2)
        self.dc_conv3 = conv(128, 128, kernel_size=3, stride=1, padding=4, dilation=4)
        self.dc_conv4 = conv(128, 96, kernel_size=3, stride=1, padding=8, dilation=8)
        self.dc_conv5 = conv(96, 64, kernel_size=3, stride=1, padding=16, dilation=16)
        self.dc_conv6 = conv(64, 32, kernel_size=3, stride=1, padding=1, dilation=1)
        self.dc_conv7 = predict_flow(32)

        # self.upfeat5 = deconv()

        self.deform6 = deformable_conv(196, 196)
        self.deform5 = deformable_conv(128, 128)
        self.deform4 = deformable_conv(96, 96)
        self.deform3 = deformable_conv(64, 64)
        self.deform2 = deformable_conv(32, 32)

    def corr(self, f1, f2):
        corr = self.corr_fn(f1, f2)
        corr = corr.view(corr.shape[0], -1, corr.shape[3], corr.shape[4])
        corr = corr / f2.shape[1]
        return corr

    def warp(self, x, flo):
        """
        warp an image/tensor (im2) back to im1, according to the optical flow
        x: [B, C, H, W] (im2)
        flo: [B, 2, H, W] flow
        """
        B, C, H, W = x.size()
        # mesh grid
        xx = torch.arange(0, W, dtype=x.dtype, device=x.device).view(1, -1).repeat(H, 1)
        yy = torch.arange(0, H, dtype=x.dtype, device=x.device).view(-1, 1).repeat(1, W)
        xx = xx.view(1, 1, H, W).repeat(B, 1, 1, 1)
        yy = yy.view(1, 1, H, W).repeat(B, 1, 1, 1)
        grid = torch.cat((xx, yy), 1)

        if x.is_cuda:
            grid = grid.cuda()
        # vgrid = Variable(grid) + flo
        vgrid = Variable(grid) + torch.flip(flo, [1])

        # scale grid to [-1,1]
        vgrid[:, 0, :, :] = 2.0 * vgrid[:, 0, :, :].clone() / max(W - 1, 1) - 1.0
        vgrid[:, 1, :, :] = 2.0 * vgrid[:, 1, :, :].clone() / max(H - 1, 1) - 1.0

        # vgrid = vgrid.permute(0,2,3,1)
        vgrid = vgrid.permute(0, 2, 3, 1).clamp(-1.1, 1.1)
        output = nn.functional.grid_sample(x, vgrid, align_corners=True)
        mask = torch.autograd.Variable(torch.ones(x.size())).cuda()
        mask = nn.functional.grid_sample(mask, vgrid, align_corners=True)

        # if W==128:
        # np.save('mask.npy', mask.cpu().data.numpy())
        # np.save('warp.npy', output.cpu().data.numpy())

        mask[mask < 0.9999] = 0
        mask[mask > 0] = 1

        return output * mask

    def forward(self, inputs):
        bgr_mean = rearrange(inputs["images"], "b n c h w -> b c (n h w)").mean(2)[
            :, None, :, None, None
        ]
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-bgr_mean,
            bgr_mult=1.0,
            bgr_to_rgb=True,
            resize_mode="interpolation",
            interpolation_mode="bilinear",
            interpolation_align_corners=False,
        )
        orig_images = inputs["images"]
        inputs["images"] = images
        mfns_out = self.MaskFlownet_S(inputs)
        srcs = mfns_out["srcs"]
        c1s, c2s, flows, c30, c40 = srcs
        c11, c12, c13, c14, c15, c16 = c1s
        c21, c22, c23, c24, c25, c26 = c2s

        c31 = self.conv1z(self.conv1y(self.conv1x(c30)))
        c32 = self.conv2z(self.conv2y(self.conv2x(c31)))
        c33 = self.conv3z(self.conv3y(self.conv3x(c32)))
        c34 = self.conv4z(self.conv4y(self.conv4x(c33)))
        c35 = self.conv5z(self.conv5y(self.conv5x(c34)))
        c36 = self.conv6z(self.conv6y(self.conv6x(c35)))

        c41 = self.conv1z(self.conv1y(self.conv1x(c40)))
        c42 = self.conv2z(self.conv2y(self.conv2x(c41)))
        c43 = self.conv3z(self.conv3y(self.conv3x(c42)))
        c44 = self.conv4z(self.conv4y(self.conv4x(c43)))
        c45 = self.conv5z(self.conv5y(self.conv5x(c44)))
        c46 = self.conv6z(self.conv6y(self.conv6x(c45)))

        flow6 = flows[0]

        warp6u = (flow6 * self.scale / self.strides[0]).unsqueeze(1)
        warp6u = torch.repeat_interleave(warp6u, 9, 1)
        S1, S2, S3, S4, S5 = warp6u.shape
        warp6u = warp6u.view(S1, S2 * S3, S4, S5)
        warp6u = self.deform6(c26, warp6u)
        warp6u = self.leakyRELU(warp6u)
        corr6u = self.leakyRELU(self.corr(c16, warp6u))
        warp6v = c46
        corr6v = self.leakyRELU(self.corr(c36, warp6v))
        x = torch.cat((corr6u, corr6v, flow6), 1)
        x = torch.cat((self.conv6_0(x), x), 1)
        x = torch.cat((self.conv6_1(x), x), 1)
        x = torch.cat((self.conv6_2(x), x), 1)
        x = torch.cat((self.conv6_3(x), x), 1)
        x = torch.cat((self.conv6_4(x), x), 1)
        flow6 = flow6 + self.pred_flow6(x)

        feat5 = self.leakyRELU(self.upfeat5(x))
        flow5 = Upsample(flow6, 2)
        warp5u = (flow5 * self.scale / self.strides[1]).unsqueeze(1)
        warp5u = torch.repeat_interleave(warp5u, 9, 1)
        S1, S2, S3, S4, S5 = warp5u.shape
        warp5u = warp5u.view(S1, S2 * S3, S4, S5)
        warp5u = self.deform5(c25, warp5u)
        warp5u = self.leakyRELU(warp5u)
        corr5u = self.leakyRELU(self.corr(c15, warp5u))
        warp5v = c45
        corr5v = self.leakyRELU(self.corr(c35, warp5v))
        x = torch.cat((c15, feat5, corr5u, corr5v, flow5, flows[1]), 1)
        x = torch.cat((self.conv5_0(x), x), 1)
        x = torch.cat((self.conv5_1(x), x), 1)
        x = torch.cat((self.conv5_2(x), x), 1)
        x = torch.cat((self.conv5_3(x), x), 1)
        x = torch.cat((self.conv5_4(x), x), 1)
        flow5 = flow5 + self.pred_flow5(x)

        feat4 = self.leakyRELU(self.upfeat4(x))
        flow4 = Upsample(flow5, 2)
        warp4u = (flow4 * self.scale / self.strides[2]).unsqueeze(1)
        warp4u = torch.repeat_interleave(warp4u, 9, 1)
        S1, S2, S3, S4, S5 = warp4u.shape
        warp4u = warp4u.view(S1, S2 * S3, S4, S5)
        warp4u = self.deform4(c24, warp4u)
        warp4u = self.leakyRELU(warp4u)
        corr4u = self.leakyRELU(self.corr(c14, warp4u))
        warp4v = c44
        corr4v = self.leakyRELU(self.corr(c34, warp4v))
        x = torch.cat((c14, feat4, corr4u, corr4v, flow4, flows[2]), 1)
        x = torch.cat((self.conv4_0(x), x), 1)
        x = torch.cat((self.conv4_1(x), x), 1)
        x = torch.cat((self.conv4_2(x), x), 1)
        x = torch.cat((self.conv4_3(x), x), 1)
        x = torch.cat((self.conv4_4(x), x), 1)
        flow4 = flow4 + self.pred_flow4(x)

        feat3 = self.leakyRELU(self.upfeat3(x))
        flow3 = Upsample(flow4, 2)
        warp3u = (flow3 * self.scale / self.strides[3]).unsqueeze(1)
        warp3u = torch.repeat_interleave(warp3u, 9, 1)
        S1, S2, S3, S4, S5 = warp3u.shape
        warp3u = warp3u.view(S1, S2 * S3, S4, S5)
        warp3u = self.deform3(c23, warp3u)
        warp3u = self.leakyRELU(warp3u)
        corr3u = self.leakyRELU(self.corr(c13, warp3u))
        warp3v = c43
        corr3v = self.leakyRELU(self.corr(c33, warp3v))
        x = torch.cat((c13, feat3, corr3u, corr3v, flow3, flows[3]), 1)
        x = torch.cat((self.conv3_0(x), x), 1)
        x = torch.cat((self.conv3_1(x), x), 1)
        x = torch.cat((self.conv3_2(x), x), 1)
        x = torch.cat((self.conv3_3(x), x), 1)
        x = torch.cat((self.conv3_4(x), x), 1)
        flow3 = flow3 + self.pred_flow3(x)

        feat2 = self.leakyRELU(self.upfeat2(x))
        flow2 = Upsample(flow3, 2)
        warp2u = (flow2 * self.scale / self.strides[4]).unsqueeze(1)
        warp2u = torch.repeat_interleave(warp2u, 9, 1)
        S1, S2, S3, S4, S5 = warp2u.shape
        warp2u = warp2u.view(S1, S2 * S3, S4, S5)
        warp2u = self.deform2(c22, warp2u)
        warp2u = self.leakyRELU(warp2u)
        corr2u = self.leakyRELU(self.corr(c12, warp2u))
        warp2v = c42
        corr2v = self.leakyRELU(self.corr(c32, warp2v))
        x = torch.cat((c12, feat2, corr2u, corr2v, flow2, flows[4]), 1)
        x = torch.cat((self.conv2_0(x), x), 1)
        x = torch.cat((self.conv2_1(x), x), 1)
        x = torch.cat((self.conv2_2(x), x), 1)
        x = torch.cat((self.conv2_3(x), x), 1)
        x = torch.cat((self.conv2_4(x), x), 1)
        flow2 = flow2 + self.pred_flow2(x)

        x = self.dc_conv4(self.dc_conv3(self.dc_conv2(self.dc_conv1(x))))
        flow2 = flow2 + self.dc_conv7(self.dc_conv6(self.dc_conv5(x)))

        preds = [
            flow.flip(1) * self.scale for flow in [flow6, flow5, flow4, flow3, flow2]
        ]
        visuals = []
        visuals.append(flow2[:, :1])

        flow_up = F.interpolate(
            preds[-1],
            size=images.shape[-2:],
            mode="bilinear",
            align_corners=False,
        )
        flow_up = self.postprocess_predictions(flow_up, image_resizer, is_flow=True)

        output = {
            "flows": flow_up[:, None],
            "visuals": visuals[-1][:, None],
        }
        if self.training:
            output["flow_preds"] = preds

        inputs["images"] = orig_images

        return output


class EpeLoss(nn.Module):
    def __init__(self, eps=0):
        super(EpeLoss, self).__init__()
        self.eps = eps

    def forward(self, pred, label):
        loss = ((pred - label).pow(2).sum(1) + self.eps).sqrt()
        return loss.view(loss.shape[0], -1).mean(1)


class EpeLossWithMask(nn.Module):
    def __init__(self, eps=1e-8, q=None):
        super(EpeLossWithMask, self).__init__()
        self.eps = eps
        self.q = q

    def forward(self, pred, label, mask):
        if self.q is not None:
            loss = ((pred - label).abs().sum(1) + self.eps) ** self.q
        else:
            loss = ((pred - label).pow(2).sum(1) + self.eps).sqrt()
        loss = loss * mask.squeeze(1)
        loss = loss.view(loss.shape[0], -1).sum(1) / mask.view(mask.shape[0], -1).sum(1)
        return loss


class MultiscaleEpe(nn.Module):
    def __init__(self, scales, weights, match, eps=1e-8, q=None):
        super(MultiscaleEpe, self).__init__()

        self.scales = scales
        self.weights = weights
        self.match = match
        self.eps = eps
        self.q = q

    def forward(self, output, target):
        flow = target["flows"][:, 0]
        mask = target["valids"][:, 0]
        predictions = output["flow_preds"]
        losses = 0
        if self.match == "upsampling":
            for p, w, s in zip(predictions, self.weights, self.scales):
                losses += (
                    EpeLossWithMask(eps=self.eps, q=self.q)(Upsample(p, s), flow, mask)
                    * w
                )
        elif self.match == "downsampling":
            for p, w, s in zip(predictions, self.weights, self.scales):
                losses += (
                    EpeLossWithMask(eps=self.eps, q=self.q)(
                        p, Downsample(flow, s), Downsample(mask, s)
                    )
                    * w
                )
        else:
            raise NotImplementedError
        return losses


@register_model
@trainable
class maskflownet_s(MaskFlownet_S):
    pass


@register_model
@trainable
class maskflownet(MaskFlownet):
    pass



# =============================================================================
# Copyright 2024 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

# TensorRT conversion code comes from the tutorial:
# https://pytorch.org/TensorRT/tutorials/_rendered_examples/dynamo/torch_compile_resnet_example.html


import sys
from argparse import ArgumentParser
from pathlib import Path
import time

import cv2 as cv
import numpy as np
import torch
import torch_tensorrt

this_dir = Path(__file__).parent.resolve()
sys.path.insert(0, str(this_dir.parent.parent.parent))

from ptlflow import get_model, load_checkpoint
from ptlflow.utils import flow_utils
from ptlflow.utils.lightning.ptlflow_cli import PTLFlowCLI
from ptlflow.utils.registry import RegisteredModel


def _init_parser() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "--ckpt_path",
        type=str,
        default=None,
        help="Path to the checkpoint to be loaded. It can also be one of the following names: {chairs, things, sintel, kitti}, in which case the respective pretrained checkpoint will be downloaded.",
    )
    parser.add_argument(
        "--image_paths",
        type=str,
        nargs=2,
        default=[
            str(this_dir / "image_samples" / "000000_10.png"),
            str(this_dir / "image_samples" / "000000_11.png"),
        ],
        help="Path to two images to estimate the optical flow with the TensorRT model.",
    )
    parser.add_argument(
        "--output_path",
        type=str,
        default=".",
        help="Path to the directory where the predictions will be saved.",
    )
    parser.add_argument(
        "--input_size",
        type=int,
        nargs=2,
        default=[384, 1280],
        help="Size of the input image.",
    )
    return parser


def compile_engine_and_infer(args):
    # Initialize model with half precision and sample inputs
    model = load_model(args).half().eval().to("cuda")
    images = [
        torch.from_numpy(load_images(args.image_paths, args.input_size))
        .contiguous()
        .half()
        .to("cuda")
    ]

    num_tries = 11
    total_time_orig = 0.0
    for i in range(num_tries):
        torch.cuda.synchronize()
        start = time.perf_counter()
        model(images[0])
        torch.cuda.synchronize()
        end = time.perf_counter()
        if i > 0:
            total_time_orig += end - start

    # Enabled precision for TensorRT optimization
    enabled_precisions = {torch.half}

    # Whether to print verbose logs
    debug = True

    # Workspace size for TensorRT
    workspace_size = 20 << 30

    # Maximum number of TRT Engines
    # (Lower value allows more graph segmentation)
    min_block_size = 7

    # Operations to Run in Torch, regardless of converter support
    torch_executed_ops = {}

    # Build and compile the model with torch.compile, using Torch-TensorRT backend
    compiled_model = torch_tensorrt.compile(
        model,
        ir="torch_compile",
        inputs=images,
        enabled_precisions=enabled_precisions,
        debug=debug,
        workspace_size=workspace_size,
        min_block_size=min_block_size,
        torch_executed_ops=torch_executed_ops,
    )

    total_time_optimized = 0.0
    for i in range(num_tries):
        torch.cuda.synchronize()
        start = time.perf_counter()
        flow_pred = compiled_model(*images)
        torch.cuda.synchronize()
        end = time.perf_counter()
        if i > 0:
            total_time_optimized += end - start

    try:
        torch_tensorrt.save(compiled_model, f"{args.model}.tc", inputs=images)
        print(f"Saving compiled model to {args.model}.tc")
        compiled_model = torch_tensorrt.load(f"{args.model}.tc")
        print(f"Loading compiled model from {args.model}.tc")
    except Exception as e:
        print("WARNING: The compiled model was not saved due to the error:")
        print(e)

    print(f"Model: {args.model}. Average time of {num_tries - 1} runs:")
    print(f"Time (original): {(1000 * total_time_orig / (num_tries - 1)):.2f} ms.")
    print(f"Time (compiled): {(1000 * total_time_optimized / (num_tries - 1)):.2f} ms.")

    flow_pred_npy = flow_pred[0].permute(1, 2, 0).detach().cpu().numpy()

    output_dir = Path(args.output_path)
    output_dir.mkdir(parents=True, exist_ok=True)

    flo_output_path = output_dir / f"flow_pred.flo"
    flow_utils.flow_write(flo_output_path, flow_pred_npy)
    print(f"Saved flow prediction to: {flo_output_path}")

    viz_output_path = output_dir / f"flow_pred_viz.png"
    flow_viz = flow_utils.flow_to_rgb(flow_pred_npy)
    cv.imwrite(str(viz_output_path), cv.cvtColor(flow_viz, cv.COLOR_RGB2BGR))
    print(f"Saved flow prediction visualization to: {viz_output_path}")

    # Finally, we use Torch utilities to clean up the workspace
    torch._dynamo.reset()


def load_images(image_paths, input_size):
    images = [cv.imread(p) for p in image_paths]
    images = [cv.resize(im, input_size[::-1]) for im in images]
    images = np.stack(images)
    images = images.transpose(0, 3, 1, 2)[None]
    images = images.astype(np.float32) / 255.0
    return images


def load_model(cfg):
    ckpt_path = cfg.ckpt_path
    cfg.ckpt_path = None  # This is to avoid letting get_model to load the ckpt
    model = get_model(cfg.model_name, args=cfg)
    # Since we set fuse_next1d_weights to True in the model,
    # the ckpt 1D layers also need to be fused before loading
    ckpt = load_checkpoint(ckpt_path, model.__class__)
    state_dict = fuse_checkpoint_next1d_layers(ckpt["state_dict"])
    model.load_state_dict(state_dict, strict=True)
    return model


def fuse_checkpoint_next1d_layers(state_dict):
    fused_sd = {}
    hv_pairs = {}
    for name, param in state_dict.items():
        if name.endswith("weight_h") or name.endswith("weight_v"):
            name_prefix = name[: -(len("weight_h") + 1)]
            orientation = name[-1]
            if name_prefix not in hv_pairs:
                hv_pairs[name_prefix] = {}
            hv_pairs[name_prefix][orientation] = param
        else:
            fused_sd[name] = param

    for name_prefix, param_pairs in hv_pairs.items():
        weight = torch.einsum("cijk,cimj->cimk", param_pairs["h"], param_pairs["v"])
        fused_sd[f"{name_prefix}.weight"] = weight
    return fused_sd


if __name__ == "__main__":
    parser = _init_parser()

    cli = PTLFlowCLI(
        model_class=RegisteredModel,
        subclass_mode_model=True,
        parser_kwargs={"parents": [parser]},
        run=False,
        parse_only=False,
        auto_configure_optimizers=False,
    )

    cfg = cli.config
    cfg.model.init_args.corr_mode = "allpairs"
    cfg.model.init_args.fuse_next1d_weights = True
    cfg.model.init_args.simple_io = True
    cfg.model_name = cfg.model.class_path.split(".")[-1]

    compile_engine_and_infer(cfg)



import torch
import torch.nn.functional as F


class ConvBlock(torch.nn.Module):
    def __init__(self, in_planes, out_planes, kernel_size, stride, padding):
        super(ConvBlock, self).__init__()

        self.conv1 = torch.nn.Conv2d(
            in_planes,
            out_planes,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            padding_mode="zeros",
            bias=False,
        )

        self.conv2 = torch.nn.Conv2d(
            out_planes, out_planes, kernel_size=3, stride=1, padding=1, bias=False
        )

        self.relu = torch.nn.LeakyReLU(negative_slope=0.1, inplace=False)

        self.norm = torch.nn.BatchNorm2d(out_planes, eps=1e-06, affine=False)

        # self.dropout = torch.nn.Dropout(p=0.1)

    def forward(self, x):
        # x = self.dropout(x)

        x1 = self.relu(self.conv1(x))
        x2 = self.relu(self.conv2(x1))

        return self.norm(x1 + x2)


class DownDimBlock(torch.nn.Module):
    def __init__(self, in_planes, out_planes):
        super(DownDimBlock, self).__init__()

        self.relu = torch.nn.LeakyReLU(negative_slope=0.1, inplace=False)

        self.conv_block = ConvBlock(
            in_planes, out_planes, kernel_size=1, stride=1, padding=0
        )

    def forward(self, x):
        return self.conv_block(self.relu(x))


class CNNEncoder(torch.nn.Module):
    def __init__(self, feature_dim):
        super(CNNEncoder, self).__init__()

        # self.conv0 = torch.nn.Conv2d(3, 8, kernel_size=3, stride=1, padding=1, padding_mode='zeros', bias=False) # rgb2gray
        # self.norm0 = torch.nn.BatchNorm2d(16, eps=1e-06, affine=False)

        self.block1_1 = ConvBlock(
            3, feature_dim, kernel_size=8, stride=8, padding=0
        )  # 1/1

        self.block1_2 = ConvBlock(
            3, feature_dim, kernel_size=8, stride=4, padding=2
        )  # 1/2

        self.block1_3 = ConvBlock(
            3, feature_dim, kernel_size=8, stride=2, padding=3
        )  # 1/4

        self.block1_4 = ConvBlock(
            3, feature_dim, kernel_size=7, stride=1, padding=3
        )  # 1/8

        self.block1_dd = DownDimBlock(feature_dim * 4, feature_dim)  # pick features
        self.block1_ds = ConvBlock(
            feature_dim, feature_dim, kernel_size=2, stride=2, padding=0
        )

        self.block2 = ConvBlock(
            3, feature_dim, kernel_size=7, stride=1, padding=3
        )  # 1/16
        self.block2_dd = DownDimBlock(feature_dim * 2, feature_dim)  # pick features

    def init_pos(self, batch_size, height, width, dtype, device):
        ys, xs = torch.meshgrid(
            torch.arange(height, dtype=dtype, device=device),
            torch.arange(width, dtype=dtype, device=device),
            indexing="ij",
        )
        ys = ys / (height - 1)
        xs = xs / (width - 1)
        pos = torch.stack([ys, xs])
        return pos[None].repeat(batch_size, 1, 1, 1)

    def init_pos_12(self, batch_size, height, width, dtype, device):
        self.pos_1 = self.init_pos(
            batch_size, height, width, dtype=dtype, device=device
        )
        self.pos_2 = self.init_pos(
            batch_size, height // 2, width // 2, dtype=dtype, device=device
        )

    def forward(self, img):
        b = img.shape[0]

        # x = self.relu(self.norm0(self.conv0(x)))

        x1_1 = self.block1_1(img)

        img = F.avg_pool2d(img, kernel_size=2, stride=2)

        x1_2 = self.block1_2(img)

        img = F.avg_pool2d(img, kernel_size=2, stride=2)

        x1_3 = self.block1_3(img)

        img = F.avg_pool2d(img, kernel_size=2, stride=2)

        x1_4 = self.block1_4(img)

        x1 = torch.cat([x1_1, x1_2, x1_3, x1_4], dim=1)
        x1 = self.block1_dd(x1)

        img = F.avg_pool2d(img, kernel_size=2, stride=2)

        x2 = self.block2(img)

        x2 = torch.cat([self.block1_ds(x1), x2], dim=1)
        x2 = self.block2_dd(x2)

        x1 = torch.cat([x1, self.pos_1], dim=1)
        x2 = torch.cat([x2, self.pos_2], dim=1)

        # x2 = self.self_attn(x2, x2)

        return [x1, x2]



import torch
import torch.nn.functional as F
import math

from . import utils


def bilinear_sample(img, coords):
    """Wrapper for grid_sample, uses pixel coordinates"""
    H, W = img.shape[-2:]
    xgrid, ygrid = coords.split([1, 1], dim=-1)
    xgrid = 2 * xgrid / (W - 1) - 1
    ygrid = 2 * ygrid / (H - 1) - 1

    grid = torch.cat([xgrid, ygrid], dim=-1)

    with torch.backends.cudnn.flags(enabled=False):
        img = F.grid_sample(img, grid, align_corners=True)

    return img


class CorrBlock:
    def __init__(self, radius, levels):
        self.radius = radius
        self.levels = levels

    def init_bhwd(self, batch_size, height, width, device, amp):
        xy_range = torch.linspace(
            -self.radius,
            self.radius,
            2 * self.radius + 1,
            dtype=torch.half if amp else torch.float,
            device=device,
        )

        delta = torch.stack(torch.meshgrid(xy_range, xy_range, indexing="ij"), axis=-1)
        delta = delta.view(1, 2 * self.radius + 1, 2 * self.radius + 1, 2)

        self.grid = utils.coords_grid(batch_size, height, width, device, amp)
        self.delta = delta.repeat(batch_size * height * width, 1, 1, 1)

    def __call__(self, corr_pyramid, flow):
        b, _, h, w = flow.shape

        coords = (self.grid + flow).permute(0, 2, 3, 1)
        coords = coords.reshape(b * h * w, 1, 1, 2)

        out_list = []

        for level, corr in enumerate(corr_pyramid):
            curr_coords = coords / 2**level + self.delta
            corr = bilinear_sample(corr, curr_coords)
            corr = corr.view(b, h, w, -1)
            out_list.append(corr)

        out = torch.cat(out_list, dim=-1)

        return out.permute(0, 3, 1, 2).contiguous()

    def init_corr_pyr(self, feature0, feature1):
        b, c, h, w = feature0.shape
        feature0 = feature0.view(b, c, h * w)
        feature1 = feature1.view(b, c, h * w)

        corr = torch.matmul(feature0.transpose(1, 2), feature1)
        corr = corr.view(b * h * w, 1, h, w) / math.sqrt(c)

        corr_pyramid = [corr]
        for i in range(self.levels - 1):
            corr = F.avg_pool2d(corr, kernel_size=2, stride=2)
            corr_pyramid.append(corr)

        return corr_pyramid



import torch
import torch.nn as nn
import torch.nn.functional as F

from .update import GMAUpdateBlock
from ..encoders import twins_svt_large
from .cnn import BasicEncoder
from .corr import CorrBlock, OLCorrBlock, AlternateCorrBlock
from ...utils.utils import bilinear_sampler, coords_grid, upflow8
from .gma import Attention, Aggregate
from .sk import SKUpdateBlock6_Deep_nopoolres_AllDecoder
from .sk2 import SKUpdateBlock6_Deep_nopoolres_AllDecoder2

from torchvision.utils import save_image

autocast = torch.cuda.amp.autocast


class BOFNet(nn.Module):
    def __init__(self, cfg):
        super().__init__()
        self.cfg = cfg

        self.hidden_dim = hdim = 128
        self.context_dim = cdim = 128

        cfg.corr_radius = 4
        cfg.corr_levels = 4

        # feature network, context network, and update block
        if cfg.cnet == "twins":
            print("[Using twins as context encoder]")
            self.cnet = twins_svt_large(pretrained=self.cfg.pretrain)
        elif cfg.cnet == "basicencoder":
            print("[Using basicencoder as context encoder]")
            self.cnet = BasicEncoder(output_dim=256, norm_fn="instance")

        if cfg.fnet == "twins":
            print("[Using twins as feature encoder]")
            self.fnet = twins_svt_large(pretrained=self.cfg.pretrain)
        elif cfg.fnet == "basicencoder":
            print("[Using basicencoder as feature encoder]")
            self.fnet = BasicEncoder(output_dim=256, norm_fn="instance")

        if self.cfg.gma == "GMA":
            print("[Using GMA]")
            self.update_block = GMAUpdateBlock(self.cfg, hidden_dim=128)
        elif self.cfg.gma == "GMA-SK":
            print("[Using GMA-SK]")
            self.cfg.cost_heads_num = 1
            self.update_block = SKUpdateBlock6_Deep_nopoolres_AllDecoder(
                args=self.cfg, hidden_dim=128
            )
        elif self.cfg.gma == "GMA-SK2":
            print("[Using GMA-SK2]")
            self.cfg.cost_heads_num = 1
            self.update_block = SKUpdateBlock6_Deep_nopoolres_AllDecoder2(
                args=self.cfg, hidden_dim=128
            )

        print("[Using corr_fn {}]".format(self.cfg.corr_fn))

        self.att = Attention(
            args=self.cfg, dim=128, heads=1, max_pos_size=160, dim_head=128
        )

    def initialize_flow(self, img):
        """Flow is represented as difference between two coordinate grids flow = coords1 - coords0"""
        N, C, H, W = img.shape
        coords0 = coords_grid(N, H // 8, W // 8).to(img.device)
        coords1 = coords_grid(N, H // 8, W // 8).to(img.device)

        # optical flow computed as difference: flow = coords1 - coords0
        return coords0, coords1

    def upsample_flow(self, flow, mask):
        """Upsample flow field [H/8, W/8, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 8, 8, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(8 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 8 * H, 8 * W)

    def forward(self, images, data={}, flow_init=None):
        B, N, _, H, W = images.shape

        images = 2 * (images / 255.0) - 1.0

        hdim = self.hidden_dim
        cdim = self.context_dim

        with autocast(enabled=self.cfg.mixed_precision):
            fmaps = self.fnet(images.reshape(B * N, 3, H, W)).reshape(
                B, N, -1, H // 8, W // 8
            )
        fmap1 = fmaps[:, 0, ...]
        fmap2 = fmaps[:, 1, ...]
        fmap3 = fmaps[:, 2, ...]

        if self.cfg.corr_fn == "efficient":
            corr_fn_21 = AlternateCorrBlock(fmap2, fmap1, radius=self.cfg.corr_radius)
            corr_fn_23 = AlternateCorrBlock(fmap2, fmap3, radius=self.cfg.corr_radius)
        else:
            corr_fn_21 = CorrBlock(
                fmap2,
                fmap1,
                num_levels=self.cfg.corr_levels,
                radius=self.cfg.corr_radius,
            )
            corr_fn_23 = CorrBlock(
                fmap2,
                fmap3,
                num_levels=self.cfg.corr_levels,
                radius=self.cfg.corr_radius,
            )

        with autocast(enabled=self.cfg.mixed_precision):
            cnet = self.cnet(images[:, 1, ...])
            net, inp = torch.split(cnet, [hdim, cdim], dim=1)
            net = torch.tanh(net)
            inp = torch.relu(inp)
            attention = self.att(inp)

        coords0_21, coords1_21 = self.initialize_flow(images[:, 0, ...])
        coords0_23, coords1_23 = self.initialize_flow(images[:, 0, ...])

        flow_predictions = []
        for itr in range(self.cfg.decoder_depth):
            coords1_21 = coords1_21.detach()
            coords1_23 = coords1_23.detach()

            corr21 = corr_fn_21(coords1_21)
            corr23 = corr_fn_23(coords1_23)
            corr = torch.cat([corr23, corr21], dim=1)

            flow21 = coords1_21 - coords0_21
            flow23 = coords1_23 - coords0_23
            flow = torch.cat([flow23, flow21], dim=1)

            with autocast(enabled=self.cfg.mixed_precision):
                net, up_mask, delta_flow = self.update_block(
                    net, inp, corr, flow, attention
                )

            up_mask_21, up_mask_23 = torch.split(up_mask, [64 * 9, 64 * 9], dim=1)

            coords1_23 = coords1_23 + delta_flow[:, 0:2, ...]
            coords1_21 = coords1_21 + delta_flow[:, 2:4, ...]

            # upsample predictions
            flow_up_23 = self.upsample_flow(coords1_23 - coords0_23, up_mask_23)
            flow_up_21 = self.upsample_flow(coords1_21 - coords0_21, up_mask_21)

            flow_predictions.append(torch.stack([flow_up_23, flow_up_21], dim=1))

        if self.training:
            return flow_predictions
        else:
            return flow_predictions[-1], torch.stack(
                [coords1_23 - coords0_23, coords1_21 - coords0_21], dim=1
            )



import torch


class KeyValueMemoryStore:
    """
    Works for key/value pairs type storage
    e.g., working and long-term memory
    """

    """
    An object group is created when new objects enter the video
    Objects in the same group share the same temporal extent
    i.e., objects initialized in the same frame are in the same group
    For DAVIS/interactive, there is only one object group
    For YouTubeVOS, there can be multiple object groups
    """

    def __init__(self, count_usage: bool):
        self.count_usage = count_usage

        # keys/value are stored in a single tensor
        self.k = None
        self.v = None

        # shrinkage and selection are also single tensors
        self.s = self.e = None

        # usage
        if self.count_usage:
            self.use_count = self.life_count = None

    def add(self, key, value, shrinkage=None, selection=None):
        new_count = torch.zeros(
            (key.shape[0], 1, key.shape[2]), device=key.device, dtype=key.dtype
        )
        new_life = (
            torch.zeros(
                (key.shape[0], 1, key.shape[2]), device=key.device, dtype=key.dtype
            )
            + 1e-7
        )

        # add the key
        if self.k is None:
            self.k = key
            self.s = shrinkage
            self.e = selection
            self.v = value
            if self.count_usage:
                self.use_count = new_count
                self.life_count = new_life
        else:
            self.k = torch.cat([self.k, key], -1)
            self.v = torch.cat([self.v, value], -1)
            if shrinkage is not None:
                self.s = torch.cat([self.s, shrinkage], -1)
            if selection is not None:
                self.e = torch.cat([self.e, selection], -1)
            if self.count_usage:
                self.use_count = torch.cat([self.use_count, new_count], -1)
                self.life_count = torch.cat([self.life_count, new_life], -1)

    def update_usage(self, usage):
        # increase all life count by 1
        # increase use of indexed elements
        if not self.count_usage:
            return

        self.use_count += usage.view_as(self.use_count)
        self.life_count += 1

    def sieve_by_range(self, start: int, end: int):
        # keep only the elements *outside* of this range (with some boundary conditions)
        # i.e., concat (a[:start], a[end:])
        # min_size is only used for values, we do not sieve values under this size
        # (because they are not consolidated)

        if end == 0:
            # negative 0 would not work as the end index!
            self.k = self.k[:, :, :start]
            self.v = self.v[:, :, :start]
            if self.count_usage:
                self.use_count = self.use_count[:, :, :start]
                self.life_count = self.life_count[:, :, :start]
            if self.s is not None:
                self.s = self.s[:, :, :start]
            if self.e is not None:
                self.e = self.e[:, :, :start]
        else:
            self.k = torch.cat([self.k[:, :, :start], self.k[:, :, end:]], -1)
            self.v = torch.cat([self.v[:, :, :start], self.v[:, :, end:]], -1)
            if self.count_usage:
                self.use_count = torch.cat(
                    [self.use_count[:, :, :start], self.use_count[:, :, end:]], -1
                )
                self.life_count = torch.cat(
                    [self.life_count[:, :, :start], self.life_count[:, :, end:]], -1
                )
            if self.s is not None:
                self.s = torch.cat([self.s[:, :, :start], self.s[:, :, end:]], -1)
            if self.e is not None:
                self.e = torch.cat([self.e[:, :, :start], self.e[:, :, end:]], -1)

    def remove_obsolete_features(self, max_size: int):
        # normalize with life duration
        usage = self.get_usage().flatten()

        values, _ = torch.topk(
            usage, k=(self.size - max_size), largest=False, sorted=True
        )
        survived = usage > values[-1]

        self.k = self.k[:, :, survived]
        self.v = self.v[:, :, survived]
        self.s = self.s[:, :, survived] if self.s is not None else None
        # Long-term memory does not store ek so this should not be needed
        self.e = self.e[:, :, survived] if self.e is not None else None

        self.use_count = self.use_count[:, :, survived]
        self.life_count = self.life_count[:, :, survived]

    def get_usage(self):
        # return normalized usage
        if not self.count_usage:
            raise RuntimeError("I did not count usage!")
        else:
            usage = self.use_count / self.life_count
            return usage

    def get_all_sliced(self, start: int, end: int):
        # return k, sk, ek, usage in order, sliced by start and end

        if end == 0:
            # negative 0 would not work as the end index!
            k = self.k[:, :, start:]
            sk = self.s[:, :, start:] if self.s is not None else None
            ek = self.e[:, :, start:] if self.e is not None else None
            usage = self.get_usage()[:, :, start:]
        else:
            k = self.k[:, :, start:end]
            sk = self.s[:, :, start:end] if self.s is not None else None
            ek = self.e[:, :, start:end] if self.e is not None else None
            usage = self.get_usage()[:, :, start:end]

        return k, sk, ek, usage

    def get_v_size(self):
        return self.v.shape[-1]

    def engaged(self):
        return self.k is not None

    @property
    def size(self):
        if self.k is None:
            return 0
        else:
            return self.k.shape[-1]

    @property
    def key(self):
        return self.k

    @property
    def value(self):
        return self.v

    @property
    def shrinkage(self):
        return self.s

    @property
    def selection(self):
        return self.e



# =============================================================================
# Copyright 2024 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import math
from typing import Tuple

import torch.nn as nn

from .local_timm.norm import LayerNorm2d
from .local_timm.weight_init import trunc_normal_
from .next1d import NeXt1DStage


class NeXt1DEncoder(nn.Module):
    def __init__(
        self,
        max_pyr_range: Tuple[int, int],
        stem_stride: int,
        num_recurrent_layers: int,
        hidden_chs: int,
        out_chs: int,
        norm_layer=LayerNorm2d,
        mlp_ratio: float = 4,
        depth: int = 2,
        fuse_next1d_weights: bool = False,
    ):
        super(NeXt1DEncoder, self).__init__()
        self.max_pyr_range = max_pyr_range
        self.stem_stride = stem_stride
        self.num_recurrent_layers = num_recurrent_layers

        self.stem = self._make_stem(hidden_chs, stem_stride, norm_layer=norm_layer)
        self.rec_stage = self._make_stage(
            hidden_chs,
            mlp_ratio=mlp_ratio,
            norm_layer=norm_layer,
            depth=depth,
            fuse_next1d_weights=fuse_next1d_weights,
        )
        self.out_layer = self._make_out_layer(hidden_chs, out_chs)

    def _make_stem(self, hidden_chs: int, stem_stride: int, norm_layer):
        conv = nn.Conv2d(3, hidden_chs, kernel_size=7, stride=stem_stride, padding=3)
        return nn.Sequential(conv, norm_layer(hidden_chs))

    def _make_stage(
        self, hidden_chs: int, norm_layer, mlp_ratio, depth, fuse_next1d_weights
    ):
        return NeXt1DStage(
            hidden_chs,
            hidden_chs,
            stride=2,
            depth=depth,
            norm_layer=norm_layer,
            mlp_ratio=mlp_ratio,
            fuse_next1d_weights=fuse_next1d_weights,
        )

    def _make_out_layer(self, hidden_chs: int, out_chs: int):
        return nn.Conv2d(hidden_chs, out_chs, kernel_size=1)

    def _init_weights(self):
        for module in self.modules():
            if isinstance(module, nn.Conv2d):
                trunc_normal_(module.weight, std=0.02)
                if module.bias is not None:
                    nn.init.zeros_(module.bias)

    def forward(self, x):
        x_pyramid = []
        curr_stride = 1
        iters = self.num_recurrent_layers + 2 - int(math.log2(self.stem_stride))
        for i in range(iters):
            if i == 0:
                x = self.stem(x)
                curr_stride *= self.stem_stride
            else:
                x = self.rec_stage(x).contiguous()
                curr_stride *= 2

            if curr_stride >= self.max_pyr_range[0]:
                x_pyramid.append(x)

        for i, x in enumerate(x_pyramid[::-1]):
            x = self.out_layer(x)
            x_pyramid[i] = x

        return x_pyramid



# =============================================================================
# Copyright 2024 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Adapted from RAFT: https://github.com/princeton-vl/RAFT
# =============================================================================

import math

from loguru import logger
import torch
import torch.nn as nn
import torch.nn.functional as F

from ptlflow.utils.registry import register_model, trainable, ptlflow_trained
from ptlflow.utils.utils import forward_interpolate_batch
from .pwc_modules import rescale_flow
from .update import UpdateBlock
from .corr import get_corr_block
from .local_timm.norm import LayerNorm2d
from ..base_model.base_model import BaseModel

from .next1d_encoder import NeXt1DEncoder
from .next1d import NeXt1DStage

try:
    import alt_cuda_corr
except:
    alt_cuda_corr = None


class SequenceLoss(nn.Module):
    def __init__(self, gamma: float, max_flow: float):
        super().__init__()
        self.gamma = gamma
        self.max_flow = max_flow

    def forward(self, outputs, inputs):
        """Loss function defined over sequence of flow predictions"""

        flow_preds = outputs["flow_preds"]
        flow_gt = inputs["flows"][:, 0]
        valid = inputs["valids"][:, 0]

        n_predictions = len(flow_preds)
        flow_loss = 0.0

        # exlude invalid pixels and extremely large diplacements
        mag = torch.sum(flow_gt**2, dim=1, keepdim=True).sqrt()
        valid = (valid >= 0.5) & (mag < self.max_flow)

        for i in range(n_predictions):
            i_weight = self.gamma ** (n_predictions - i - 1)
            i_loss = (flow_preds[i] - flow_gt).abs()
            flow_loss += i_weight * (valid * i_loss).mean()

        return flow_loss


class RAPIDFlow(BaseModel):
    pretrained_checkpoints = {
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/rapidflow-chairs-9c8c182a.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/rapidflow-things-0377c8fa.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/rapidflow-sintel-89a21262.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/rapidflow-kitti-2561329f.ckpt",
    }

    def __init__(
        self,
        pyramid_ranges: tuple[int, int] = (32, 8),
        iters: int = 12,
        corr_mode: str = "allpairs",
        corr_levels: int = 1,
        corr_range: int = 4,
        enc_hidden_chs: int = 64,
        enc_out_chs: int = 128,
        enc_stem_stride: int = 4,
        enc_mlp_ratio: float = 4.0,
        enc_depth: int = 4,
        dec_net_chs: int = 64,
        dec_inp_chs: int = 64,
        dec_motion_chs: int = 128,
        dec_depth: int = 2,
        dec_mlp_ratio: float = 4.0,
        use_upsample_mask: bool = True,
        fuse_next1d_weights: bool = False,
        simple_io: bool = False,
        gamma: float = 0.8,
        max_flow: float = 400,
        **kwargs,
    ) -> None:
        num_recurrent_layers = int(math.log2(max(pyramid_ranges))) - 1
        super().__init__(
            output_stride=int(2 ** (num_recurrent_layers + 1)),
            loss_fn=SequenceLoss(gamma, max_flow),
            **kwargs,
        )

        self.pyramid_ranges = pyramid_ranges
        self.iters = iters
        self.corr_mode = corr_mode
        self.corr_levels = corr_levels
        self.corr_range = corr_range
        self.enc_hidden_chs = enc_hidden_chs
        self.enc_out_chs = enc_out_chs
        self.enc_stem_stride = enc_stem_stride
        self.enc_mlp_ratio = enc_mlp_ratio
        self.enc_depth = enc_depth
        self.dec_net_chs = dec_net_chs
        self.dec_inp_chs = dec_inp_chs
        self.dec_motion_chs = dec_motion_chs
        self.dec_depth = dec_depth
        self.dec_mlp_ratio = dec_mlp_ratio
        self.use_upsample_mask = use_upsample_mask
        self.fuse_next1d_weights = fuse_next1d_weights
        self.simple_io = simple_io

        for v in self.pyramid_ranges:
            assert (
                v > 1
            ), f"--pyramid_ranges values must be larger than 1, but found {v}"
            log_res = math.log2(v)
            assert (log_res) - int(
                log_res
            ) < 1e-3, f"--pyramid_ranges values must be powers of 2, but found {v}"
        num_recurrent_layers = int(math.log2(max(self.pyramid_ranges))) - 1

        self.pyramid_levels = [
            num_recurrent_layers + 1 - int(math.log2(v)) for v in self.pyramid_ranges
        ]

        max_pyr_range = (min(self.pyramid_ranges), max(self.pyramid_ranges))
        self.fnet = NeXt1DEncoder(
            max_pyr_range=max_pyr_range,
            stem_stride=self.enc_stem_stride,
            num_recurrent_layers=num_recurrent_layers,
            hidden_chs=self.enc_hidden_chs,
            out_chs=self.enc_out_chs,
            mlp_ratio=self.enc_mlp_ratio,
            norm_layer=LayerNorm2d,
            depth=self.enc_depth,
            fuse_next1d_weights=self.fuse_next1d_weights,
        )

        self.cnet = NeXt1DEncoder(
            max_pyr_range=max_pyr_range,
            stem_stride=self.enc_stem_stride,
            num_recurrent_layers=num_recurrent_layers,
            hidden_chs=self.enc_hidden_chs,
            out_chs=self.enc_out_chs,
            mlp_ratio=self.enc_mlp_ratio,
            norm_layer=LayerNorm2d,
            depth=self.enc_depth,
            fuse_next1d_weights=self.fuse_next1d_weights,
        )

        self.dim_corr = (self.corr_range * 2 + 1) ** 2 * self.corr_levels

        self.update_block = UpdateBlock(
            pyramid_ranges=pyramid_ranges,
            corr_levels=corr_levels,
            corr_range=corr_range,
            dec_net_chs=dec_net_chs,
            dec_inp_chs=dec_inp_chs,
            dec_motion_chs=dec_motion_chs,
            dec_depth=dec_depth,
            dec_mlp_ratio=dec_mlp_ratio,
            fuse_next1d_weights=fuse_next1d_weights,
            use_upsample_mask=use_upsample_mask,
        )

        self.upnet_layer = nn.Sequential(
            nn.Conv2d(2 * self.dec_net_chs, self.dec_net_chs, 1),
            nn.ReLU(inplace=True),
            NeXt1DStage(
                self.dec_net_chs,
                self.dec_net_chs,
                stride=1,
                depth=2,
                mlp_ratio=self.dec_mlp_ratio,
                norm_layer=LayerNorm2d,
                fuse_next1d_weights=self.fuse_next1d_weights,
            ),
        )

        self.has_trained_on_ptlflow = True

        if self.corr_mode == "local" and alt_cuda_corr is None:
            logger.warning(
                "!!! alt_cuda_corr is not compiled! The slower IterativeCorrBlock will be used instead !!!"
            )

    def coords_grid(self, batch, ht, wd, dtype, device):
        coords = torch.meshgrid(
            torch.arange(ht, dtype=dtype, device=device),
            torch.arange(wd, dtype=dtype, device=device),
            indexing="ij",
        )
        coords = torch.stack(coords[::-1], dim=0).to(dtype=dtype)
        return coords[None].repeat(batch, 1, 1, 1)

    def upsample_flow(self, flow, mask, factor):
        """Upsample flow field [H/f, W/f, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, factor, factor, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, factor * H, factor * W)

    def forward(self, inputs):
        if self.simple_io:
            images = inputs
        else:
            images = inputs["images"]
        images, image_resizer = self.preprocess_images(
            images,
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=False,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=True,
        )

        x1_raw = images[:, 0]
        x2_raw = images[:, 1]
        b, _, height_im, width_im = x1_raw.size()

        x_pyramid = self.fnet(torch.cat([x1_raw, x2_raw], 0))
        x1_pyramid = [x[:b] for x in x_pyramid]
        x2_pyramid = [x[b:] for x in x_pyramid]

        # outputs
        flows = []

        cnet_pyramid = self.cnet(x1_raw)

        pred_stride = min(self.pyramid_ranges)

        start_level, output_level = self.pyramid_levels
        pass_pyramid1 = x1_pyramid[start_level : output_level + 1]
        pass_pyramid2 = x2_pyramid[start_level : output_level + 1]
        pass_pyramid_cnet = cnet_pyramid[start_level : output_level + 1]

        iters_per_level = [
            int(math.ceil(float(self.iters) / (output_level - start_level + 1)))
        ] * (output_level - start_level + 1)

        # init
        (
            b_size,
            _,
            h_x1,
            w_x1,
        ) = pass_pyramid1[0].size()
        init_device = pass_pyramid1[0].device

        if (
            not self.simple_io
            and "prev_flows" in inputs
            and inputs["prev_flows"] is not None
        ):
            flow = F.interpolate(
                inputs["prev_flows"][:, 0],
                [pass_pyramid1[0].shape[-2], pass_pyramid1[0].shape[-1]],
                mode="bilinear",
                align_corners=True,
            )
            flow = rescale_flow(flow, width_im, height_im, to_local=True)
            flow = forward_interpolate_batch(flow)
        else:
            flow = torch.zeros(
                b_size, 2, h_x1, w_x1, dtype=x1_raw.dtype, device=init_device
            )

        net = None
        for l, (x1, x2, cnet) in enumerate(
            zip(pass_pyramid1, pass_pyramid2, pass_pyramid_cnet)
        ):
            coords0 = self.coords_grid(
                x1.shape[0], x1.shape[2], x1.shape[3], x1.dtype, x1.device
            )

            corr_fn = get_corr_block(
                x1,
                x2,
                self.corr_levels,
                self.corr_range,
                alternate_corr=self.corr_mode == "local",
            )

            net_tmp, inp = torch.split(
                cnet, [self.dec_net_chs, self.dec_inp_chs], dim=1
            )
            inp = torch.relu(inp)

            if net is None:
                net = torch.tanh(net_tmp)
            else:
                net = F.interpolate(
                    net,
                    [x1.shape[-2], x1.shape[-1]],
                    mode="bilinear",
                    align_corners=True,
                )

                net_skip = torch.tanh(net_tmp)
                gate = torch.sigmoid(
                    self.upnet_layer(torch.cat([net, net_skip], dim=1))
                )
                net = gate * net + (1.0 - gate) * net_skip

            if l > 0:
                flow = rescale_flow(flow, x1.shape[-1], x1.shape[-2], to_local=False)
                flow = F.interpolate(
                    flow,
                    [x1.shape[-2], x1.shape[-1]],
                    mode="bilinear",
                    align_corners=True,
                )

            for k in range(iters_per_level[l]):
                flow = flow.detach()

                # correlation
                out_corr = corr_fn(coords0 + flow)

                get_mask = self.training or (
                    l == (output_level - start_level) and k == (iters_per_level[l] - 1)
                )
                flow_res, net, mask = self.update_block(
                    net, inp, out_corr, flow, get_mask=get_mask
                )
                flow = flow + flow_res

                out_flow = rescale_flow(flow, width_im, height_im, to_local=False)
                if self.training:
                    if mask is not None and l == (output_level - start_level):
                        if self.simple_io:
                            # Just copied the code from self.upsample_flow to here.
                            # For some reason, TensorRT backend does not compile when calling the function
                            N, _, H, W = out_flow.shape
                            mask = mask.view(N, 1, 9, pred_stride, pred_stride, H, W)
                            mask = torch.softmax(mask, dim=2)

                            up_flow = F.unfold(flow, [3, 3], padding=1)
                            up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

                            up_flow = torch.sum(mask * up_flow, dim=2)
                            up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
                            up_flow = up_flow.reshape(
                                N, 2, pred_stride * H, pred_stride * W
                            )
                            out_flow = up_flow
                        else:
                            out_flow = self.upsample_flow(out_flow, mask, pred_stride)
                    else:
                        out_flow = F.interpolate(
                            out_flow,
                            [x1_raw.shape[-2], x1_raw.shape[-1]],
                            mode="bilinear",
                            align_corners=True,
                        )
                elif l == (output_level - start_level) and k == (
                    iters_per_level[l] - 1
                ):
                    if mask is not None:
                        if self.simple_io:
                            # Just copied the code from self.upsample_flow to here.
                            # For some reason, TensorRT backend does not compile when calling the function
                            N, _, H, W = out_flow.shape
                            mask = mask.view(N, 1, 9, pred_stride, pred_stride, H, W)
                            mask = torch.softmax(mask, dim=2)

                            up_flow = F.unfold(flow, [3, 3], padding=1)
                            up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

                            up_flow = torch.sum(mask * up_flow, dim=2)
                            up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
                            up_flow = up_flow.reshape(
                                N, 2, pred_stride * H, pred_stride * W
                            )
                            out_flow = up_flow
                        else:
                            out_flow = self.upsample_flow(out_flow, mask, pred_stride)
                    else:
                        out_flow = F.interpolate(
                            out_flow,
                            [x1_raw.shape[-2], x1_raw.shape[-1]],
                            mode="bilinear",
                            align_corners=True,
                        )
                out_flow = self.postprocess_predictions(
                    out_flow, image_resizer, is_flow=True
                )
                flows.append(out_flow)

        if self.simple_io:
            return flows[-1]
        else:
            outputs = {}
            outputs["flows"] = flows[-1][:, None]
            if self.training:
                outputs["flow_preds"] = flows
            return outputs


class RAPIDFlow_it1(RAPIDFlow):
    def __init__(
        self,
        pyramid_ranges: tuple[int, int] = (32, 32),
        iters: int = 1,
        corr_mode: str = "allpairs",
        corr_levels: int = 1,
        corr_range: int = 4,
        enc_hidden_chs: int = 64,
        enc_out_chs: int = 128,
        enc_stem_stride: int = 4,
        enc_mlp_ratio: float = 4,
        enc_depth: int = 4,
        dec_net_chs: int = 64,
        dec_inp_chs: int = 64,
        dec_motion_chs: int = 128,
        dec_depth: int = 2,
        dec_mlp_ratio: float = 4,
        use_upsample_mask: bool = False,
        fuse_next1d_weights: bool = False,
        simple_io: bool = False,
        gamma: float = 0.8,
        max_flow: float = 400,
        **kwargs,
    ) -> None:
        super().__init__(
            pyramid_ranges,
            iters,
            corr_mode,
            corr_levels,
            corr_range,
            enc_hidden_chs,
            enc_out_chs,
            enc_stem_stride,
            enc_mlp_ratio,
            enc_depth,
            dec_net_chs,
            dec_inp_chs,
            dec_motion_chs,
            dec_depth,
            dec_mlp_ratio,
            use_upsample_mask,
            fuse_next1d_weights,
            simple_io,
            gamma,
            max_flow,
            **kwargs,
        )


class RAPIDFlow_it2(RAPIDFlow):
    def __init__(
        self,
        pyramid_ranges: tuple[int, int] = (32, 16),
        iters: int = 2,
        corr_mode: str = "allpairs",
        corr_levels: int = 1,
        corr_range: int = 4,
        enc_hidden_chs: int = 64,
        enc_out_chs: int = 128,
        enc_stem_stride: int = 4,
        enc_mlp_ratio: float = 4,
        enc_depth: int = 4,
        dec_net_chs: int = 64,
        dec_inp_chs: int = 64,
        dec_motion_chs: int = 128,
        dec_depth: int = 2,
        dec_mlp_ratio: float = 4,
        use_upsample_mask: bool = False,
        fuse_next1d_weights: bool = False,
        simple_io: bool = False,
        gamma: float = 0.8,
        max_flow: float = 400,
        **kwargs,
    ) -> None:
        super().__init__(
            pyramid_ranges,
            iters,
            corr_mode,
            corr_levels,
            corr_range,
            enc_hidden_chs,
            enc_out_chs,
            enc_stem_stride,
            enc_mlp_ratio,
            enc_depth,
            dec_net_chs,
            dec_inp_chs,
            dec_motion_chs,
            dec_depth,
            dec_mlp_ratio,
            use_upsample_mask,
            fuse_next1d_weights,
            simple_io,
            gamma,
            max_flow,
            **kwargs,
        )


class RAPIDFlow_it3(RAPIDFlow):
    def __init__(
        self,
        pyramid_ranges: tuple[int, int] = (32, 8),
        iters: int = 3,
        corr_mode: str = "allpairs",
        corr_levels: int = 1,
        corr_range: int = 4,
        enc_hidden_chs: int = 64,
        enc_out_chs: int = 128,
        enc_stem_stride: int = 4,
        enc_mlp_ratio: float = 4,
        enc_depth: int = 4,
        dec_net_chs: int = 64,
        dec_inp_chs: int = 64,
        dec_motion_chs: int = 128,
        dec_depth: int = 2,
        dec_mlp_ratio: float = 4,
        use_upsample_mask: bool = True,
        fuse_next1d_weights: bool = False,
        simple_io: bool = False,
        gamma: float = 0.8,
        max_flow: float = 400,
        **kwargs,
    ) -> None:
        super().__init__(
            pyramid_ranges,
            iters,
            corr_mode,
            corr_levels,
            corr_range,
            enc_hidden_chs,
            enc_out_chs,
            enc_stem_stride,
            enc_mlp_ratio,
            enc_depth,
            dec_net_chs,
            dec_inp_chs,
            dec_motion_chs,
            dec_depth,
            dec_mlp_ratio,
            use_upsample_mask,
            fuse_next1d_weights,
            simple_io,
            gamma,
            max_flow,
            **kwargs,
        )


class RAPIDFlow_it6(RAPIDFlow):
    def __init__(
        self,
        pyramid_ranges: tuple[int, int] = (32, 8),
        iters: int = 6,
        corr_mode: str = "allpairs",
        corr_levels: int = 1,
        corr_range: int = 4,
        enc_hidden_chs: int = 64,
        enc_out_chs: int = 128,
        enc_stem_stride: int = 4,
        enc_mlp_ratio: float = 4,
        enc_depth: int = 4,
        dec_net_chs: int = 64,
        dec_inp_chs: int = 64,
        dec_motion_chs: int = 128,
        dec_depth: int = 2,
        dec_mlp_ratio: float = 4,
        use_upsample_mask: bool = True,
        fuse_next1d_weights: bool = False,
        simple_io: bool = False,
        gamma: float = 0.8,
        max_flow: float = 400,
        **kwargs,
    ) -> None:
        super().__init__(
            pyramid_ranges,
            iters,
            corr_mode,
            corr_levels,
            corr_range,
            enc_hidden_chs,
            enc_out_chs,
            enc_stem_stride,
            enc_mlp_ratio,
            enc_depth,
            dec_net_chs,
            dec_inp_chs,
            dec_motion_chs,
            dec_depth,
            dec_mlp_ratio,
            use_upsample_mask,
            fuse_next1d_weights,
            simple_io,
            gamma,
            max_flow,
            **kwargs,
        )


@register_model
@trainable
@ptlflow_trained
class rapidflow(RAPIDFlow):
    pass


@register_model
@trainable
@ptlflow_trained
class rapidflow_it1(RAPIDFlow_it1):
    pass


@register_model
@trainable
@ptlflow_trained
class rapidflow_it2(RAPIDFlow_it2):
    pass


@register_model
@trainable
@ptlflow_trained
class rapidflow_it3(RAPIDFlow_it3):
    pass


@register_model
@trainable
@ptlflow_trained
class rapidflow_it6(RAPIDFlow_it6):
    pass



import torch
from torch import nn
import torch.nn.functional as F
import numpy as np


class ConvNextLayer(nn.Module):
    def __init__(self, dim, depth=4):
        super().__init__()
        self.net = nn.Sequential(*[ConvNextBlock(dim=dim) for j in range(depth)])

    def forward(self, x):
        return self.net(x)

    def compute_params(self):
        num = 0
        for param in self.parameters():
            num += np.prod(param.size())

        return num


class ConvNextBlock(nn.Module):
    r"""ConvNeXt Block. There are two equivalent implementations:
    (1) DwConv -> LayerNorm (channels_first) -> 1x1 Conv -> GELU -> 1x1 Conv; all in (N, C, H, W)
    (2) DwConv -> Permute to (N, H, W, C); LayerNorm (channels_last) -> Linear -> GELU -> Linear; Permute back
    We use (2) as we find it slightly faster in PyTorch

    Args:
        dim (int): Number of input channels.
        drop_path (float): Stochastic depth rate. Default: 0.0
        layer_scale_init_value (float): Init value for Layer Scale. Default: 1e-6.
    """

    def __init__(self, dim, layer_scale_init_value=1e-6):
        super().__init__()
        self.dwconv = nn.Conv2d(
            dim, dim, kernel_size=7, padding=3, groups=dim
        )  # depthwise conv
        self.norm = LayerNorm(dim, eps=1e-6)
        self.pwconv1 = nn.Linear(
            dim, 4 * dim
        )  # pointwise/1x1 convs, implemented with linear layers
        self.act = nn.GELU()
        self.pwconv2 = nn.Linear(4 * dim, dim)
        self.gamma = (
            nn.Parameter(layer_scale_init_value * torch.ones((dim)), requires_grad=True)
            if layer_scale_init_value > 0
            else None
        )
        # self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        # print(f"conv next layer")

    def forward(self, x):
        input = x
        x = self.dwconv(x)
        x = x.permute(0, 2, 3, 1)  # (N, C, H, W) -> (N, H, W, C)
        x = self.norm(x)
        x = self.pwconv1(x)
        x = self.act(x)
        x = self.pwconv2(x)
        if self.gamma is not None:
            x = self.gamma * x
        x = x.permute(0, 3, 1, 2)  # (N, H, W, C) -> (N, C, H, W)

        x = input + x
        return x


class LayerNorm(nn.Module):
    r"""LayerNorm that supports two data formats: channels_last (default) or channels_first.
    The ordering of the dimensions in the inputs. channels_last corresponds to inputs with
    shape (batch_size, height, width, channels) while channels_first corresponds to inputs
    with shape (batch_size, channels, height, width).
    """

    def __init__(self, normalized_shape, eps=1e-6, data_format="channels_last"):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(normalized_shape))
        self.bias = nn.Parameter(torch.zeros(normalized_shape))
        self.eps = eps
        self.data_format = data_format
        if self.data_format not in ["channels_last", "channels_first"]:
            raise NotImplementedError
        self.normalized_shape = (normalized_shape,)

    def forward(self, x):
        if self.data_format == "channels_last":
            return F.layer_norm(
                x, self.normalized_shape, self.weight, self.bias, self.eps
            )
        elif self.data_format == "channels_first":
            u = x.mean(1, keepdim=True)
            s = (x - u).pow(2).mean(1, keepdim=True)
            x = (x - u) / torch.sqrt(s + self.eps)
            x = self.weight[:, None, None] * x + self.bias[:, None, None]
            return x



import math
import torch
from typing import Optional


def get_similarity(mk, ms, qk, qe):
    # used for training/inference and memory reading/memory potentiation
    # mk: B x CK x [N]    - Memory keys
    # ms: B x  1 x [N]    - Memory shrinkage
    # qk: B x CK x [HW/P] - Query keys
    # qe: B x CK x [HW/P] - Query selection
    # Dimensions in [] are flattened
    CK = mk.shape[1]
    mk = mk.flatten(start_dim=2)
    ms = ms.flatten(start_dim=1).unsqueeze(2) if ms is not None else None
    qk = qk.flatten(start_dim=2)
    qe = qe.flatten(start_dim=2) if qe is not None else None

    if qe is not None:
        # See appendix for derivation
        # or you can just trust me ヽ(ー_ー )ノ
        mk = mk.transpose(1, 2)
        a_sq = mk.pow(2) @ qe
        two_ab = 2 * (mk @ (qk * qe))
        b_sq = (qe * qk.pow(2)).sum(1, keepdim=True)
        similarity = -a_sq + two_ab - b_sq
    else:
        # similar to STCN if we don't have the selection term
        a_sq = mk.pow(2).sum(1).unsqueeze(2)
        two_ab = 2 * (mk.transpose(1, 2) @ qk)
        similarity = -a_sq + two_ab

    if ms is not None:
        similarity = similarity * ms / math.sqrt(CK)  # B*N*HW
    else:
        similarity = similarity / math.sqrt(CK)  # B*N*HW

    return similarity


def do_softmax(
    similarity, top_k: Optional[int] = None, inplace=False, return_usage=False
):
    # normalize similarity with top-k softmax
    # similarity: B x N x [HW/P]
    # use inplace with care
    if top_k is not None:
        values, indices = torch.topk(similarity, k=top_k, dim=1)

        x_exp = values.exp_()
        x_exp /= torch.sum(x_exp, dim=1, keepdim=True)
        if inplace:
            similarity.zero_().scatter_(1, indices, x_exp)  # B*N*HW
            affinity = similarity
        else:
            affinity = torch.zeros_like(similarity).scatter_(
                1, indices, x_exp
            )  # B*N*HW
    else:
        maxes = torch.max(similarity, dim=1, keepdim=True)[0]
        x_exp = torch.exp(similarity - maxes)
        x_exp_sum = torch.sum(x_exp, dim=1, keepdim=True)
        affinity = x_exp / x_exp_sum
        indices = None

    if return_usage:
        return affinity, affinity.sum(dim=2)

    return affinity


def get_affinity(mk, ms, qk, qe):
    # shorthand used in training with no top-k
    similarity = get_similarity(mk, ms, qk, qe)
    affinity = do_softmax(similarity)
    return affinity


def readout(affinity, mv):
    B, CV, T, H, W = mv.shape

    mo = mv.view(B, CV, T * H * W)
    mem = torch.bmm(mo, affinity)
    mem = mem.view(B, CV, H, W)

    return mem



"""Various utility functions."""

# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import logging
import math
from argparse import ArgumentParser
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
import torch
import torch.nn.functional as F
import yaml

import ptlflow
from ptlflow.utils.external.raft import InputPadder as _InputPadder, forward_interpolate


class InputPadder(_InputPadder):
    """Pads images such that dimensions are divisible by stride.

    This is just a wrapper for ptlflow.utils.external.raft.InputPadder.
    """

    def __init__(
        self,
        dims: Sequence[int],
        stride: int,
        size: Optional[Tuple[int, int]] = None,
        two_side_pad: bool = True,
        pad_mode: str = "replicate",
        pad_value: float = 0.0,
    ) -> None:
        """Initialize InputPadder.

        Parameters
        ----------
        dims : Sequence[int]
            The shape of the original input. It must have at least two elements. It is assumed that the last two dimensions
            are (height, width).
        stride : int
            The number to compute the amount of padding. The padding will be applied so that the input size is divisible
            by stride.
        size : Optional[Tuple[int, int]], optional
            The desired size after scaling defined as (height, width). If not provided, then scale_factor will be used instead.
        two_side_pad : bool, default True
            If True, half of the padding goes to left/top and the rest to right/bottom. Otherwise, all the padding goes to the bottom right.
        pad_mode : str, default "replicate"
            How to pad the input. Must be one of the values accepted by the 'mode' argument of torch.nn.functional.pad.
        pad_value : float, default 0.0
            Used if pad_mode == "constant". The value to fill in the padded area.
        """
        super().__init__(
            dims,
            stride=stride,
            size=size,
            two_side_pad=two_side_pad,
            pad_mode=pad_mode,
            pad_value=pad_value,
        )
        if size is None:
            self.tgt_size = (
                int(math.ceil(float(dims[-2]) / stride)) * stride,
                int(math.ceil(float(dims[-1]) / stride)) * stride,
            )
        else:
            self.tgt_size = size

    def fill(self, x):
        return self.pad(x)

    def unfill(self, x):
        if x.shape[-2] == self.tgt_size[0] and x.shape[-1] == self.tgt_size[1]:
            x = self.unpad(x)
        return x


class InputScaler(object):
    """Scale 2D torch.Tensor input to a target size, and then rescale it back to the original size."""

    def __init__(
        self,
        orig_shape: Tuple[int, int],
        stride: Optional[int] = None,
        size: Optional[Tuple[int, int]] = None,
        scale_factor: Optional[float] = 1.0,
        interpolation_mode: str = "bilinear",
        interpolation_align_corners: bool = False,
    ) -> None:
        """Initialize InputScaler.

        Parameters
        ----------
        orig_shape : Tuple[int, int]
            The shape of the input tensor before the scale. I.e., the shape to which it will be rescaled back.
        stride : Optional[int], optional
            If provided, the input will be resized to the closest larger multiple of stride.
        size : Optional[Tuple[int, int]], optional
            The desired size after scaling defined as (height, width). If not provided, then scale_factor will be used instead.
        scale_factor : Optional[float], default 1.0
            This value is only used if stride and size are None. The multiplier that will be applied to the original shape to scale
            the input.
        interpolation_mode : str, default 'bilinear'
            How to perform the interpolation. It must be a value accepted by the 'mode' argument from
            torch.nn.functional.interpolate function.
        interpolation_align_corners : bool, default False
            Whether the interpolation keep the corners aligned. As defined in torch.nn.functional.interpolate.

        See Also
        --------
        torch.nn.functional.interpolate : The function used to scale the inputs.
        """
        super().__init__()
        self.orig_height, self.orig_width = orig_shape[-2:]
        if stride is not None:
            assert size is None, "only stride OR size can be provided, NOT BOTH."
            self.tgt_height = int(math.ceil(float(self.orig_height) / stride)) * stride
            self.tgt_width = int(math.ceil(float(self.orig_width) / stride)) * stride
        elif size is not None:
            assert stride is None, "only stride OR size can be provided, NOT BOTH."
            self.tgt_height, self.tgt_width = size
        else:
            self.tgt_height = int(self.orig_height * scale_factor)
            self.tgt_width = int(self.orig_width * scale_factor)

        self.interpolation_mode = interpolation_mode
        self.interpolation_align_corners = interpolation_align_corners

    def fill(self, x: torch.Tensor, is_flow: bool = False) -> torch.Tensor:
        """Scale the input to the target size specified during initialization.

        Parameters
        ----------
        x : torch.Tensor
            The input to be scaled. Its shape must be (..., C, H, W), where ... means any number of dimensions.
        is_flow : bool
            Whether the input is a flow field or not. If it is, then its values are multiplied by the rescale factor.

        Returns
        -------
        torch.Tensor
            The scaled input.
        """
        return self._scale_keep_dims(x, (self.tgt_height, self.tgt_width), is_flow)

    def unfill(self, x: torch.Tensor, is_flow: bool = False) -> torch.Tensor:
        """Scale the input to back to the original size defined during initialization.

        Parameters
        ----------
        x : torch.Tensor
            The input to be rescaled back. Its shape must be (..., C, H, W), where ... means any number of dimensions.
        is_flow : bool
            Whether the input is a flow field or not. If it is, then its values are multiplied by the rescale factor.

        Returns
        -------
        torch.Tensor
            The rescaled input.
        """
        return self._scale_keep_dims(x, (self.orig_height, self.orig_width), is_flow)

    def _scale_keep_dims(
        self, x: torch.Tensor, size: Tuple[int, int], is_flow: bool
    ) -> torch.Tensor:
        """Scale the input to a given size while keeping the other dimensions intact.

        Parameters
        ----------
        x : torch.Tensor
            The input to be rescaled back. Its shape must be (..., C, H, W), where ... means any number of dimensions.
        size : Tuple[int, int]
            The target size to scale the input.
        is_flow : bool
            Whether the input is a flow field or not. If it is, then its values are multiplied by the rescale factor.

        Returns
        -------
        torch.Tensor
            The rescaled input.
        """
        x_shape = x.shape
        x = x.view(-1, x.shape[-3], x.shape[-2], x.shape[-1])
        x = F.interpolate(
            x,
            size=size,
            mode=self.interpolation_mode,
            align_corners=self.interpolation_align_corners,
        )

        if is_flow:
            x[:, 0] = x[:, 0] * (float(x.shape[-1]) / x_shape[-1])
            x[:, 1] = x[:, 1] * (float(x.shape[-2]) / x_shape[-2])

        new_shape = list(x_shape)
        new_shape[-2], new_shape[-1] = x.shape[-2], x.shape[-1]
        x = x.view(new_shape)
        return x


def add_datasets_to_parser(
    parser: ArgumentParser, dataset_config_path: str
) -> ArgumentParser:
    """Add dataset paths as parser arguments.

    The dataset names and default paths are loaded from a yaml file.

    Parameters
    ----------
    parser : ArgumentParser
        An initialized parser, this function will add more arguments to it.
    dataset_config_path : str
        The path to the yaml file containing the dataset paths.

    Returns
    -------
    ArgumentParser
        The updated parser.
    """
    with open(dataset_config_path, "r") as f:
        dataset_paths = yaml.safe_load(f)
    for name, path in dataset_paths.items():
        parser.add_argument(
            f"--{name}_root_dir",
            type=str,
            default=path,
            help=f"Path to the root of the {name} dataset",
        )
    return parser


def config_logging() -> None:
    """Initialize logging parameters."""
    log_dir = Path("ptlflow_logs")
    log_dir.mkdir(exist_ok=True)
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s: %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        level=logging.INFO,
        handlers=[
            logging.FileHandler(log_dir / "log_run.txt"),
            logging.StreamHandler(),
        ],
    )


def count_parameters(model: torch.nn.Module) -> int:
    """Return the number of trainable parameters of a model.

    Taken from: https://discuss.pytorch.org/t/how-do-i-check-the-number-of-parameters-of-a-model/4325/7.

    Parameters
    ----------
    model : torch.nn.Module
        The model to count the parameters from.

    Returns
    -------
    int
        The number of trainable parameters of the given model.
    """
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def get_list_of_available_models_list() -> List[str]:
    """Return a list of the names of the available models.

    Returns
    -------
    list[str]
        The list with the model names.
    """
    return sorted(ptlflow.models_dict.keys())


def make_divisible(v: int, div: int) -> int:
    """Decrease a number v until it is divisible by div.

    Parameters
    ----------
    v : int
        The number to be made divisible.
    div : int
        The divisor value.

    Returns
    -------
    int
        The new value of v which is divisible by div.
    """
    return max(div, v - (v % div))


def release_gpu(tensors_dict: Dict[str, Any]) -> Dict[str, Any]:
    """Detach and move to cpu the tensors from the input dict.

    The non-tensor elements are kept intact.

    Parameters
    ----------
    tensors_dict : Dict[str, Any]
        A dictionary containing the tensors to move.

    Returns
    -------
    Dict[str, Any]
        The same dictionary, but with the tensors moved to cpu.
    """
    for k, v in tensors_dict.items():
        if isinstance(v, torch.Tensor):
            tensors_dict[k] = v.detach().cpu()
            del v
    return tensors_dict


def tensor_dict_to_numpy(
    tensor_dict: Dict[str, torch.Tensor], padder: Optional[InputPadder] = None
) -> Dict[str, np.ndarray]:
    """Convert all tensors into numpy format, changing the shape from CHW to HWC.

    If "flows" is available, then a color representation "flows_viz" is added to the outputs.

    Parameters
    ----------
    tensor_dict : dict[str, torch.Tensor]
        A dictionary with the torch.Tensor inputs/outputs of the model.
    padder: InputPadder
        Helper to unpad the images back to their original sizes.

    Returns
    -------
    dict[str, np.ndarray]
        The torch.Tensor entries from tensor_dict converted to numpy format.
    """
    npy_dict = {}
    for k, v in tensor_dict.items():
        if isinstance(v, torch.Tensor):
            v = v.detach().cpu()
            if padder is not None:
                v = padder.unpad(v)

            while len(v.shape) > 3:
                v = v[0]
            v = v.permute(1, 2, 0).numpy()
        npy_dict[k] = v
    return npy_dict


def are_shapes_compatible(
    shape1: Sequence[int],
    shape2: Sequence[int],
) -> bool:
    """Check if two tensor shapes are compatible.

    Similar to PyTorch or Numpy, two shapes are considered "compatible" if either they have the same shape or if one shape can be broadcasted into the other.
    We consider two shapes compatible if, and only if:
    1. their shapes have the same length (same number of dimension), and
    2. each dimension size is either equal or at least one of them is one.

    Parameters
    ----------
    shape1 : Sequence[int]
        The dimensions of the first shape.
    shape2 : Sequence[int]
        The dimensions of the second shape.

    Returns
    -------
    bool
        Whether the two given shapes are compatible.
    """
    if len(shape1) != len(shape2):
        return False
    for v1, v2 in zip(shape1, shape2):
        if v1 != 1 and v2 != 1 and v1 != v2:
            return False
    return True


def bgr_val_as_tensor(
    bgr_val: Union[float, Tuple[float, float, float], np.ndarray, torch.Tensor],
    reference_tensor: torch.Tensor,
    bgr_tensor_shape_position: int = -3,
) -> torch.Tensor:
    """Convert multiple types of BGR values given as input to a torch.Tensor where the BGR values are in the same position as a reference tensor.

    The bgr values can be:
    - a single number, in which case it will be repeated three times to represent BGR.
    - a tuple, list, np.ndarray, or torch.Tensor with three elements.

    The resulting tensor will have the BGR values in the same index position as the reference_tensor.
    For example, given a reference tensor with shape [B, 3, H, W] and setting bgr_tensor_shape_position == -3
    indicates that the BGR position in this reference_tensor is at shape index -3, which is equivalent to index 1.
    Given these inputs, the resulting BGR tensor will have shape [1, 3, 1, 1], and the BGR values will be at shape index 1.

    Parameters
    ----------
    bgr_val : Union[float, Tuple[float, float, float], np.ndarray, torch.Tensor]
        The BGR values to be converted into a tensor that will have a compatible shape with the input.
    reference_tensor: torch.Tensor
        The tensor with the reference shape to convert the bgr_val.
    bgr_tensor_shape_position : int, default -3
        Which position of the reference_tensor corresponds to the BGR values.
        Typical values are -1 (used in channels-last tensors) or -3 (..., CHW tensors)

    Returns
    -------
    torch.Tensor
        The bgr_val converted into a tensor with a shape compatible with reference_tensor.
    """
    is_compatible = False
    if isinstance(bgr_val, torch.Tensor):
        is_compatible = are_shapes_compatible(bgr_val.shape, reference_tensor.shape)
        assert is_compatible or (len(bgr_val.shape) == 1 and bgr_val.shape[0] == 3)
    elif isinstance(bgr_val, np.ndarray):
        is_compatible = are_shapes_compatible(bgr_val.shape, reference_tensor.shape)
        assert is_compatible or (len(bgr_val.shape) == 1 and bgr_val.shape[0] == 3)
        bgr_val = torch.from_numpy(bgr_val).to(
            dtype=reference_tensor.dtype, device=reference_tensor.device
        )
    elif isinstance(bgr_val, (tuple, list)):
        assert len(bgr_val) == 3
        bgr_val = torch.Tensor(bgr_val).to(
            dtype=reference_tensor.dtype, device=reference_tensor.device
        )
    elif isinstance(bgr_val, (int, float)):
        bgr_val = (
            torch.zeros(3, dtype=reference_tensor.dtype, device=reference_tensor.device)
            + bgr_val
        )

    if not is_compatible:
        bgr_dims = [1] * len(reference_tensor.shape)
        bgr_dims[bgr_tensor_shape_position] = 3
        bgr_val = bgr_val.reshape(bgr_dims)
    return bgr_val


def forward_interpolate_batch(prev_flow: torch.Tensor) -> torch.Tensor:
    """Apply RAFT's forward_interpolate in a batch of torch.Tensors.

    forward_interpolate in the warm start strategy where the previous flow estimation is forward projected
    and then used as initialization for the next estimation.

    Parameters
    ----------
    prev_flow : torch.Tensor
        A 4D tensor [B, 2, H, W] containing a batch of previous flow predictions.

    Returns
    -------
    torch.Tensor
        The previous flow predictions after being forward interpolated.
    """
    forward_flow = []
    for i in range(prev_flow.shape[0]):
        forward_flow.append(
            forward_interpolate(prev_flow[i]).to(
                dtype=prev_flow.dtype, device=prev_flow.device
            )
        )
    forward_flow = torch.stack(forward_flow, 0)
    return forward_flow
